import re
from collections import OrderedDict

from cucu.config import CONFIG

CONFIG.define(
    "CUCU_BROKEN_IMAGES_PAGE_CHECK",
    "enabled/disabled the broken image page checker",
    default="enabled",
)
CONFIG.define(
    "CUCU_READY_STATE_PAGE_CHECK",
    "enabled/disabled the ready state page checker",
    default="enabled",
)


def init_global_hook_variables():
    CONFIG["__CUCU_BEFORE_ALL_HOOKS"] = []
    CONFIG["__CUCU_AFTER_ALL_HOOKS"] = []

    CONFIG["__CUCU_BEFORE_SCENARIO_HOOKS"] = []
    CONFIG["__CUCU_AFTER_SCENARIO_HOOKS"] = []

    CONFIG["__CUCU_BEFORE_STEP_HOOKS"] = []
    CONFIG["__CUCU_AFTER_STEP_HOOKS"] = []

    CONFIG["__CUCU_PAGE_CHECK_HOOKS"] = OrderedDict()
    CONFIG["__CUCU_HTML_REPORT_TAG_HANDLERS"] = {}
    CONFIG["__CUCU_HTML_REPORT_SCENARIO_SUBHEADER_HANDLER"] = []

    CONFIG["__CUCU_CUSTOM_FAILURE_HANDLERS"] = []
    CONFIG["__CUCU_BEFORE_RETRY_HOOKS"] = []
    CONFIG["__CUCU_CTX"] = None


def init_scenario_hook_variables():
    CONFIG["__CUCU_BEFORE_THIS_SCENARIO_HOOKS"] = []
    CONFIG["__CUCU_AFTER_THIS_SCENARIO_HOOKS"] = []


def register_before_all_hook(hook_func):
    """
    register a before all hook that will execute once before anything is
    executed and we will pass the current behave context object as the only
    argument to the hook_func provided.
    """
    CONFIG["__CUCU_BEFORE_ALL_HOOKS"].append(hook_func)


def register_after_all_hook(hook_func):
    """
    register an after all hook that will execute once after everything is
    executed and we will pass the current behave context object as the only
    argument to the hook_func provided.
    """
    CONFIG["__CUCU_AFTER_ALL_HOOKS"].append(hook_func)


def register_before_scenario_hook(hook_func):
    """
    register a before scenario hook that will execute before every scenario has
    executed and we will pass the current behave context object as the only
    argument to the hook_func provided.
    """
    CONFIG["__CUCU_BEFORE_SCENARIO_HOOKS"].append(hook_func)


def register_after_scenario_hook(hook_func):
    """
    register an after scenario hook that will execute after every scenario has
    executed and we will pass the current behave context object as the only
    argument to the hook_func provided.
    """
    CONFIG["__CUCU_AFTER_SCENARIO_HOOKS"].append(hook_func)


def register_after_this_scenario_hook(hook_func):
    """
    register an after scenario hook that will only once after the next scenario
    has executed and we will pass the current behave context object as the only
    argument to the hook_func provided.
    """
    CONFIG["__CUCU_AFTER_THIS_SCENARIO_HOOKS"].append(hook_func)


def register_before_step_hook(hook_func):
    """
    register a before step hook that will execute before every step is executed
    and we will pass the current behave context object as the only argument to
    the hook_func provided.
    """
    CONFIG["__CUCU_BEFORE_STEP_HOOKS"].append(hook_func)


def register_after_step_hook(hook_func):
    """
    register an after step hook that will execute after every step executes and
    we will pass the current behave context object as the only argument to the
    hook_func provided.
    """
    CONFIG["__CUCU_AFTER_STEP_HOOKS"].append(hook_func)


def register_page_check_hook(name, hook_func):
    """
    register a hook to run after every "page load", a "page load" event isn't
    accurate but we try to simply check the page after every interaction such as
    clicking or navigating.

    parameters:
        name(string):        name used when logging for the page check
        hook_func(function): function called with the current
                             `cucu.browser.Browser` instance
    """
    CONFIG["__CUCU_PAGE_CHECK_HOOKS"][name] = hook_func


def register_custom_variable_handling(regex, lookup):
    """
    register a regex to match variable names on and allow the lookup
    function provided to do the handling of the resolution of the variable
    name.

    parameters:
        regex(string): regular expression to match on any config variable
                       name when doing lookups
        lookup(func): a function that accepts a variable name to return its
                      value at runtime.

                      def lookup(name):
                        return [value of the variable at runtime]
    """
    CONFIG.register_custom_variable_handling(regex, lookup)


def register_custom_tags_in_report_handling(regex, handler):
    """
    register a regex to match tag names when generating the HTML test report
    and the function provided will process the exact tag found and return
    the HTML/text to place in the report when a tag that matches the provided
    regex is found.

    parameters:
        regex(string): regular expression to match on any tag name in the HTML
                       test report.
        handler(func): a function that accepts the complete tag name (ie @foo)
                       and returns the exact value to put in the HTMl test
                       report (HTML/text).

                      def handle(tag):
                        return "<a href="...">{tag}</a>"
    """
    CONFIG["__CUCU_HTML_REPORT_TAG_HANDLERS"][re.compile(regex)] = handler


def register_custom_scenario_subheader_in_report_handling(handler):
    """
    register a handler to add HTML content under the scenario header.

    parameters:
        handler(func): a function that accepts the scenario and returns a HTML to
                       go below the scenario header

                      def handle(scenario):
                        return f"<div>this is below the {scenario['name']}'s header</div>"
    """
    CONFIG["__CUCU_HTML_REPORT_SCENARIO_SUBHEADER_HANDLER"].append(handler)


def register_custom_junit_failure_handler(handler):
    """
    register a callback function to call when there is a test failure and
    we want to augment the failure output message in the junit results
    files.

    parameters:
        handler(feature, scenario):
          where feature and scenario are the following behave model
          objects:

            * https://behave.readthedocs.io/en/stable/api.html#behave.model.Feature
            * https://behave.readthedocs.io/en/stable/api.html#behave.model.Scenario

          You must return a string that will be appended to the top of the
          failure message in the JUnit XML results files.
    """
    CONFIG["__CUCU_CUSTOM_FAILURE_HANDLERS"].append(handler)


def register_before_retry_hook(hook_func):
    """
    register an before retry hook that will execute before retrying
    cucu.utils.retry function call. The arguments passed to the provided hook
    are only the current behave context object.

    parameters:
        handler(ctx): a method that accepts the current behave context and
                      executes whatever it needs to before a retryable call is
                      about to be retried.

                      def handle(ctx):
                        ctx.do_something()
    """
    CONFIG["__CUCU_BEFORE_RETRY_HOOKS"].append(hook_func)



import sys
import threading
import traceback


class ThreadDumper(threading.Thread):
    """
    thread dumping class that can be easily stopped by calling the `.stop()`
    method.
    """

    def __init__(self, interval_min, stdout, *args, **kwargs):
        super(ThreadDumper, self).__init__(*args, **kwargs)
        self.interval_min = interval_min
        self.stdout = stdout
        self.running = False
        self.event = threading.Event()

    def stop(self):
        self.running = False
        self.event.set()

    def start(self):
        self.running = True
        super().start()

    def run(self):
        while self.running:
            self.stdout.write(f"\n{'*' * 80}\n")
            for thread in threading.enumerate():
                self.stdout.write(f"{thread}\n")
                traceback.print_stack(
                    sys._current_frames()[thread.ident], file=self.stdout
                )
                self.stdout.write("\n")
            self.stdout.write(f"{'*' * 80}\n")
            self.event.wait(self.interval_min * 60)


def start(interval_min):
    """
    start the thread dumper and return it so one can call the `.stop()` method
    on it to get it to quit

    parameters:
        interval_min(float): number of minutes between each thread stacktrace
                             dump to print

    returns:
        a ThreadDumper object that is a child of the threading.Thread class with
        the method `.stop()` to get the thread to quit
    """
    thread = ThreadDumper(interval_min, sys.stdout)
    thread.start()
    return thread



from cucu import fuzzy, helpers
from cucu.utils import (
    find_n_click_input_parent_label,
    is_element_size_zero,
    take_saw_element_screenshot,
)

from . import base_steps


def find_checkbox(ctx, name, index=0):
    """
    find a checkbox on screen using name and index provided

        * <input type="checkbox">
        * <* role="checkbox">

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired checkbox on screen
      index(str):  the index of the checkbox if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()

    element = fuzzy.find(
        ctx.browser,
        name,
        [
            'input[type="checkbox"]',
            '*[role="checkbox"]',
        ],
        index=index,
        direction=fuzzy.Direction.RIGHT_TO_LEFT,
    )

    take_saw_element_screenshot(ctx, "checkbox", name, index, element)

    return element


def is_checked(checkbox):
    """
    internal method to check a checkbox is checked
    """
    return (
        checkbox.get_attribute("checked")
        or checkbox.get_attribute("aria-checked") == "true"
    )


def is_not_checked(checkbox):
    """
    internal method to check a checkbox is not checked
    """
    return not is_checked(checkbox)


def check_checkbox(ctx, checkbox):
    """
    internal method used to check a checkbox if it is not already checked
    """
    ctx.check_browser_initialized()

    if is_checked(checkbox):
        raise AssertionError("checkbox already checked")

    if base_steps.is_disabled(checkbox):
        raise AssertionError("unable to check the checkbox, as it is disabled")

    if is_element_size_zero(checkbox):
        find_n_click_input_parent_label(ctx, checkbox)
        return

    ctx.browser.click(checkbox)


def uncheck_checkbox(ctx, checkbox):
    """
    internal method used to uncheck a checkbox if it is not already unchecked
    """
    ctx.check_browser_initialized()

    if is_not_checked(checkbox):
        raise AssertionError("checkbox already unchecked")

    if is_element_size_zero(checkbox):
        find_n_click_input_parent_label(ctx, checkbox)
        return

    ctx.browser.click(checkbox)


helpers.define_should_see_thing_with_name_steps("checkbox", find_checkbox)
helpers.define_action_on_thing_with_name_steps(
    "checkbox", "check", find_checkbox, check_checkbox
)
helpers.define_action_on_thing_with_name_steps(
    "checkbox", "uncheck", find_checkbox, uncheck_checkbox
)
helpers.define_thing_with_name_in_state_steps(
    "checkbox", "checked", find_checkbox, is_checked
)
helpers.define_thing_with_name_in_state_steps(
    "checkbox", "not checked", find_checkbox, is_not_checked
)
helpers.define_thing_with_name_in_state_steps(
    "checkbox", "disabled", find_checkbox, base_steps.is_disabled
)
helpers.define_thing_with_name_in_state_steps(
    "checkbox", "not disabled", find_checkbox, base_steps.is_not_disabled
)
helpers.define_run_steps_if_I_can_see_element_with_name_steps(
    "checkbox", find_checkbox
)
helpers.define_ensure_state_on_thing_with_name_steps(
    "checkbox", "checked", find_checkbox, is_checked, check_checkbox
)
helpers.define_ensure_state_on_thing_with_name_steps(
    "checkbox", "not checked", find_checkbox, is_not_checked, uncheck_checkbox
)



"""
various cucu utilities can be placed here and then exposed publicly through
the src/cucu/__init__.py
"""

import hashlib
import logging
import os
import pkgutil
import re
import shutil
import threading
import time
from datetime import datetime
from pathlib import Path

import humanize
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.by import By
from tabulate import DataRow, TableFormat, tabulate
from tenacity import (
    after_log,
    before_log,
    before_sleep_log,
    retry_if_not_exception_type,
    stop_after_delay,
    wait_fixed,
)
from tenacity import retry as retrying

from cucu import logger
from cucu.ansi_parser import ansi_to_html
from cucu.browser.core import Browser
from cucu.config import CONFIG

GHERKIN_TABLEFORMAT = TableFormat(
    lineabove=None,
    linebelowheader=None,
    linebetweenrows=None,
    linebelow=None,
    headerrow=DataRow("|", "|", "|"),
    datarow=DataRow("|", "|", "|"),
    padding=1,
    with_header_hide=["lineabove"],
)


class StopRetryException(Exception):
    pass


def generate_short_id(seed=None, length=7):
    """
    Generate a short character ID based on seed (defaults to current performance counter).
    """
    if seed is None:
        seed = time.perf_counter()

    return hashlib.sha256(str(seed).encode("utf-8")).hexdigest()[:length]


def format_gherkin_table(table, headings=[], prefix=""):
    formatted = tabulate(table, headings, tablefmt=GHERKIN_TABLEFORMAT)
    if prefix == "":
        return formatted

    return prefix + formatted.replace("\n", f"\n{prefix}")


#
# code below adapted from:
# https://github.com/behave/behave/blob/994dbfe30e2a372182ea613333e06f069ab97d4b/behave/runner.py#L385
# so we can have the sub steps printed in the console logs
#
def run_steps(ctx, steps_text):
    """
    run sub steps within an existing step definition but also log their output
    so that its easy to see what is happening.
    """

    # -- PREPARE: Save original ctx data for current step.
    # Needed if step definition that called this method uses .table/.text
    original_table = getattr(ctx, "table", None)
    original_text = getattr(ctx, "text", None)

    # first time a given step calls substeps we want to move the step_index
    # so it starts the followings steps the right point.
    if not ctx.current_step.has_substeps:
        ctx.step_index += 1

    ctx.current_step.has_substeps = True

    ctx.feature.parser.variant = "steps"
    steps = ctx.feature.parser.parse_steps(steps_text)

    current_step = ctx.current_step
    current_step_start_at = current_step.start_at

    # XXX: I want to get back to this and find a slightly better way to handle
    #      these substeps without mucking around with so much state in behave
    #      but for now this works correctly and existing tests work as expected.
    try:
        with ctx._use_with_behave_mode():
            for step in steps:
                for formatter in ctx._runner.formatters:
                    step.is_substep = True
                    formatter.insert_step(step, index=ctx.step_index)

                passed = step.run(ctx._runner, quiet=False, capture=False)

                if not passed:
                    if "StopRetryException" in step.error_message:
                        raise StopRetryException(step.error_message)
                    else:
                        raise AssertionError(step.error_message)

            # -- FINALLY: Restore original ctx data for current step.
            ctx.table = original_table
            ctx.text = original_text
    finally:
        ctx.current_step = current_step
        ctx.current_step.start_at = current_step_start_at

    return True


def retry(func, wait_up_to_s=None, retry_after_s=None):
    """
    utility retry function that can retry the provided `func` for the maximum
    amount of seconds specified by `wait_up_to_s` and wait the number of seconds
    specified in `retry_after_s`
    """
    if wait_up_to_s is None:
        wait_up_to_s = float(CONFIG["CUCU_STEP_WAIT_TIMEOUT_S"])

    if retry_after_s is None:
        retry_after_s = float(CONFIG["CUCU_STEP_RETRY_AFTER_S"])

    @retrying(
        stop=stop_after_delay(wait_up_to_s),
        wait=wait_fixed(retry_after_s),
        retry=retry_if_not_exception_type(StopRetryException),
        before=before_log(logger, logging.DEBUG),
        after=after_log(logger, logging.DEBUG),
        before_sleep=before_sleep_log(logger, logging.DEBUG),
        reraise=True,
    )
    def new_decorator(*args, **kwargs):
        ctx = CONFIG["__CUCU_CTX"]

        for hook in CONFIG["__CUCU_BEFORE_RETRY_HOOKS"]:
            hook(ctx)

        return func(*args, **kwargs)

    return new_decorator


def load_jquery_lib():
    """
    load jquery library
    """
    jquery_lib = pkgutil.get_data(
        "cucu", "external/jquery/jquery-3.5.1.min.js"
    )
    return jquery_lib.decode("utf8")


def text_in_current_frame(browser: Browser) -> str:
    """
    Utility to get all the visible text of the current frame.

    Args:
        browser (Browser): the browser session switched to the desired frame
    """
    script = "return window.jqCucu && jqCucu.fn.jquery;"
    jquery_version = browser.execute(script)
    if not jquery_version:
        browser.execute(load_jquery_lib())
        browser.execute("window.jqCucu = jQuery.noConflict(true);")
    text = browser.execute(
        'return jqCucu("body").children(":visible").text();'
    )
    return text


def ellipsize_filename(raw_filename):
    max_filename = 100
    new_raw_filename = normalize_filename(raw_filename)
    if len(new_raw_filename) < max_filename:
        return new_raw_filename

    ellipsis = "..."
    # save the last chars, as the ending is often important
    end_count = 40
    front_count = max_filename - (len(ellipsis) + end_count)
    ellipsized_filename = (
        new_raw_filename[:front_count]
        + ellipsis
        + new_raw_filename[-1 * end_count :]
    )
    return ellipsized_filename


def normalize_filename(raw_filename):
    normalized_filename = (
        raw_filename.replace('"', "")
        .replace("{", "")
        .replace("}", "")
        .replace("#", "")
        .replace("&", "")
    )
    return normalized_filename


def get_step_image_dir(step_index, step_name):
    """
    generate .png image file name that meets these criteria:
     - hides secrets
     - escaped
     - filename does not exceed 255 chars (OS limitation)
     - uniqueness comes from step number
    """
    escaped_step_name = CONFIG.hide_secrets(step_name).replace("/", "_")
    unabridged_dirname = f"{step_index:0>4} - {escaped_step_name}"
    dirname = ellipsize_filename(unabridged_dirname)

    return dirname


def take_saw_element_screenshot(ctx, thing, name, index, element=None):
    observed = "saw" if element else "did not see"
    prefix = "" if index == 0 else f"{humanize.ordinal(index + 1)} "

    take_screenshot(
        ctx,
        ctx.current_step.name,
        label=f'{observed} {prefix}{thing} "{name}"',
        element=element,
    )


def take_screenshot(ctx, step_name, label="", element=None):
    step = ctx.current_step
    if not hasattr(step, "screenshots"):
        step.screenshots = []

    screenshot_dir = os.path.join(
        ctx.scenario_dir, get_step_image_dir(ctx.step_index, step_name)
    )
    if not os.path.exists(screenshot_dir):
        os.mkdir(screenshot_dir)

    if len(label) > 0:
        label = CONFIG.hide_secrets(label).replace("/", "_")

    filename = f"{CONFIG['__STEP_SCREENSHOT_COUNT']:0>4} - {label}.png"
    filename = ellipsize_filename(filename)
    filepath = os.path.join(screenshot_dir, filename)

    ctx.browser.screenshot(filepath)

    # driver.get_window_size returns the size of the window the OS draws for
    # the browser, not the window the browser uses to display the DOM.
    # If we go through JavaScript, we ignore the height of the adress bar
    # and such.
    script = """
        function getDimensionsOfCurrentWindow() {
          const windowSize = {
            width: window.innerWidth,
            height: window.innerHeight,
          };
          return windowSize;
        };
        return getDimensionsOfCurrentWindow();
    """
    size = ctx.browser.execute(script)

    location = {
        "x": None,
        "y": None,
        "height": None,
        "width": None,
    }
    highlight = {}
    if element:
        location.update(element.rect)
        highlight = {
            "height_ratio": location["height"] / size["height"],
            "width_ratio": location["width"] / size["width"],
            "top_ratio": location["y"] / size["height"],
            "left_ratio": location["x"] / size["width"],
        }

    filename = os.path.split(filepath)[-1]
    if not getattr(step, "step_image_dir", None):
        step.step_image_dir = get_step_image_dir(ctx.step_index, step_name)

    screenshot_index = len(step.screenshots) + 1
    id = f"step-img-{step.step_run_id}-{screenshot_index:0>4}"

    src = Path(step.step_image_dir) / filename
    screenshot = {
        "step_name": step_name,
        "label": label or step_name,
        "location": location,
        "filepath": filepath,
        "size": size,
        "highlight": highlight,
        "index": screenshot_index,
        "html_src": str(src),
        "html_id": id,
    }
    step.screenshots.append(screenshot)

    if CONFIG["CUCU_MONITOR_PNG"]:
        shutil.copyfile(filepath, CONFIG["CUCU_MONITOR_PNG"])

    CONFIG["__STEP_SCREENSHOT_COUNT"] += 1


def find_n_click_input_parent_label(ctx, input_element):
    """
    Clicks the nearest parent <label> of an input elemnt (if input is visually hidden or size is zero).
    """
    try:
        # Find the closest ancestor <label> element
        label = input_element.find_element(By.XPATH, "ancestor::label[1]")

        if label and label.is_displayed():
            ctx.browser.click(label)
            logger.debug("Successfully clicked the parent label.")
        else:
            logger.warning("Parent label is not displayed or not found.")

    except Exception as e:
        logger.error(
            f"Click on parent label failed (possibly missing label ancestor): {e}"
        )


def is_element_size_zero(element):
    size = element.size
    return size["width"] == 0 and size["height"] == 0


class TeeStream:
    """
    A stream that writes to both a file stream and captures content in an internal buffer.
    Provides file-like accessors to read the captured content.
    """

    def __init__(self, file_stream):
        self.file_stream = file_stream
        self.string_buffer = []

    def write(self, data):
        self.file_stream.write(data)
        self.string_buffer.append(data)

    def flush(self):
        self.file_stream.flush()

    def getvalue(self):
        return "".join(self.string_buffer)

    def read(self):
        return self.getvalue()

    def clear(self):
        """Clear the internal buffer."""
        self.string_buffer = []

    @property
    def encoding(self):
        return self.file_stream.encoding


def get_iso_timestamp_with_ms():
    """
    Get the current time as an ISO 8601 formatted string with milliseconds precision.
    """
    return datetime.now().isoformat()[:-3]


def parse_iso_timestamp(iso_timestamp: (str | None)) -> datetime | None:
    """
    Parse an ISO 8601 formatted string with milliseconds precision into a datetime object.
    Returns None if the input is None.
    """
    if iso_timestamp is None:
        return None

    return datetime.fromisoformat(iso_timestamp)


def get_feature_name(filename):
    filepath = Path(filename)
    if ":" in filepath.name:
        filepath = filepath.parent / filepath.name.split(":")[0]

    text = filepath.read_text(encoding="utf8")
    lines = text.split("\n")
    for line in lines:
        if "Feature:" in line:
            feature_name = line.replace("Feature:", "").strip()
            return feature_name


def behave_filepath_to_cucu_logpath(filepath: Path, results: Path) -> Path:
    if ":" in filepath.name:
        filepath = filepath.parent / filepath.name.split(":")[0]

    if filepath.is_dir():
        log_filepath = results / "run.console.log"
    else:
        log_filepath = results / f"{get_feature_name(filepath)}.console.log"

    return log_filepath


def build_debug_output(raw: str) -> list[str]:
    ANSI_CURSOR_RE = re.compile(r"\x1b\[[0-9;]*[ABCD]")
    MULTI_NL_RE = re.compile(r"\n{3,}")

    # normalize newlines
    text = raw.replace("\r\n", "\n").replace("\r", "\n")

    # remove cursor movement junk
    text = ANSI_CURSOR_RE.sub("", text)

    # collapse insane blank lines
    text = MULTI_NL_RE.sub("\n\n", text)

    lines = []
    for line in text.split("\n"):
        # keep empty lines (spacing matters)
        if not line:
            lines.append("<br>")
            continue

        html = ansi_to_html(line)
        lines.append(html + "<br>")

    return lines


class SeleniumKeepAlive:
    """
    Manages keep-alive for Selenium sessions during long-running hooks.

    This prevents the Selenium session from timing out while API calls
    are being made in after_scenario hooks by periodically pinging the
    browser with lightweight commands.

    Example usage:
        # Start keep-alive
        keep_alive = SeleniumKeepAlive(browser, interval=30)
        keep_alive.start()

        # Do long-running operations...

        # Stop keep-alive
        keep_alive.stop()
    """

    def __init__(self, browser, interval):
        """
        Initialize the keep-alive manager.

        Args:
            browser: The selenium browser/driver instance
            interval: How often to ping the session (in seconds). Default is 30.
        """
        self.browser = browser
        self.interval = interval
        self.active = False
        self.thread = None

    def start(self):
        """
        Start the keep-alive background thread.
        This creates and starts a daemon thread that will periodically
        ping the Selenium session to prevent timeouts.
        """

        if not self.active and self.browser:
            self.active = True
            self.start_time = time.time()
            self.max_duration = int(
                CONFIG["CUCU_SELENIUM_KEEP_ALIVE_MAX_DURATION_S"]
            )
            self.thread = threading.Thread(
                target=self._keep_alive_loop, daemon=True
            )
            self.thread.start()
            logger.debug(
                f"Selenium keep-alive started (ping interval: {self.interval}s)"
            )

    def stop(self):
        """
        Stop the keep-alive background thread.
        This signals the thread to stop and waits up to 5 seconds for it to terminate.
        """
        if self.active:
            self.active = False
            if self.thread:
                self.thread.join(timeout=5)
            logger.debug("Selenium keep-alive stopped")

    def _keep_alive_loop(self):
        """
        Background thread that periodically pings Selenium to keep the session alive.

        This runs a lightweight command to prevent timeouts.
        The thread will automatically stop if the session is closed or an error occurs.
        """

        while self.active:
            if time.time() - self.start_time > self.max_duration:
                logger.warning("Keep-alive max duration reached, stopping")
                self.active = False
                break
            try:
                if self.browser:
                    # Execute a lightweight command to keep session alive
                    _ = self.browser.execute("return 1")
                    logger.debug("Keep-alive ping sent to Selenium session")
            except WebDriverException as e:
                # Session is already closed, stop the thread
                logger.warning(
                    f"Selenium session closed, stopping keep-alive: {e}"
                )
                self.active = False
                break
            except Exception as e:
                # Log but continue - don't let errors stop the keep-alive
                logger.warning(f"Keep-alive ping failed (continuing): {e}")

            # Wait before next ping
            time.sleep(self.interval)



import sys

import humanize

from cucu import logger, retry, run_steps
from cucu.config import CONFIG


class step(object):
    """
    Only to be used in this file as we're redefining the @step decorator in
    order to fix the code location of the @step() usage.

    Basically when you use any of the `define_xxx` steps below internally they
    call @step() and when that decorator executes in behave it will take the
    location of the function provided to the decorator and use that to both
    report as the location of the step's "source"
    (ie uv run behave --no-summary --format steps.doc --dry-run). Now that
    would be confusing as the step is likely defined somewhere under
    `src/cucu/steps` but the location in the behave steps output would be wrong
    and then at runtime if there's an exception throw from one of the other
    functions inside the `define_xxx` method that trace would also not originate
    from the `src/cucu/steps` location where the `define_xxx` was called.

    So this redefinition of @step will basically use a hook in the existing
    @step definition in `src/cucu/behave_tweaks.py` and provide a function
    to "fix" the inner_step function so its own code location now points back
    to the location where the `define_xxx` was called from.
    """

    def __init__(self, step_text):
        self.step_text = step_text
        frame = sys._getframe(1).f_back

        def fix_inner_step(func):
            #
            # Here is where we "fix" the code location for the function being
            # passed to the step() decorator. We are basically realigning the
            # filename and line location so it points to the original caller
            # of the `define_xxx` method. To do this alignment though we have
            # to subtract from the original file line number the number of code
            # lines in the `src/cucu/behave_tweaks.py` between the functions
            # wrapper and inner_step as that is the "actual code location" at
            # runtime and below we can only set the `firstlineno` which is the
            # "first line" of our __code__ point and by the time those N lines
            # execute the line number python would print would be N lines ahead
            # of the original location of the code we're trying to link to.
            #
            func.__code__ = func.__code__.replace(
                co_filename=frame.f_code.co_filename,
                co_firstlineno=frame.f_lineno - 3,
                co_name=frame.f_code.co_name,
            )

        self.fix_inner_step = fix_inner_step

    def __call__(self, func):
        from behave import step

        @step(self.step_text, fix_inner_step=self.fix_inner_step)
        def wrapper(*args, **kwargs):
            func(*args, **kwargs)

        return wrapper


def nth_to_ordinal(index):
    return "" if index == 0 else f'"{humanize.ordinal(index + 1)}" '


def define_should_see_thing_with_name_steps(thing, find_func, with_nth=False):
    """
    defines steps with with the following signatures:

      I should immediately see the {thing} "{name}"
      I should see the {thing} "{name}"
      I wait to see the {thing} "{name}"
      I wait up to "{seconds}" seconds to see the {thing} "{name}"

      when with_nth=True we also define:
      I should immediately see the "{nth}" {thing} "{name}"
      I should see the "{nth}" {thing} "{name}"
      I wait to see the "{nth}" {thing} "{name}"
      I wait up to "{seconds}" seconds to see the "{nth}" {thing} "{name}"

      I should immediately not see the {thing} "{name}"
      I should not see the {thing} "{name}"
      I wait to not see the {thing} "{name}"
      I wait up to "{seconds}" seconds to not see the {thing} "{name}"

      when with_nth=True we also define:
      I should immediately not see the "{nth}" {thing} "{name}"
      I should not see the "{nth}" {thing} "{name}"
      I wait to not see the "{nth}" {thing} "{name}"
      I wait up to "{seconds}" seconds to not see the "{nth}" {thing} "{name}"

    parameters:
        thing(string):       name of the thing we're creating the steps for such
                             as button, dialog, etc.
        find_func(function): function that returns the desired element:

                             def find_func(ctx, name, index=):
                                '''
                                ctx(object):  behave context object
                                name(string): name of the thing to find
                                index(int):   when there are multiple elements
                                              with the same name and you've
                                              specified with_nth=True
                                '''
        with_nth(bool):     when set to True we'll define the expanded set of
                            "nth" steps. default: False
    """

    # should see
    # undecorated def for reference below
    def base_should_see_the(ctx, thing, name, index=0):
        prefix = nth_to_ordinal(index)
        element = find_func(ctx, name, index=index)

        if element is None:
            raise AssertionError(
                f'unable to find the {prefix}{thing} "{name}"'
            )
        logger.debug(f'Success: saw {prefix}{thing} "{name}"')

    @step(f'I should immediately see the {thing} "{{name}}"')
    def should_immediately_see_the(ctx, thing, name):
        base_should_see_the(ctx, thing, name)

    @step(f'I should see the {thing} "{{name}}"')
    def should_see_the(ctx, name):
        retry(
            base_should_see_the,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, thing, name)

    @step(f'I wait to see the {thing} "{{name}}"')
    def wait_to_see_the(ctx, name):
        retry(base_should_see_the)(ctx, thing, name)

    @step(f'I wait up to "{{seconds}}" seconds to see the {thing} "{{name}}"')
    def wait_up_to_seconds_to_see_the(ctx, seconds, name):
        milliseconds = float(seconds)
        retry(base_should_see_the, wait_up_to_s=milliseconds)(ctx, thing, name)

    if with_nth:

        @step(f'I should immediately see the "{{nth:nth}}" {thing} "{{name}}"')
        def base_should_see_the_nth(ctx, nth, name):
            base_should_see_the(ctx, thing, name, index=nth)

        @step(f'I should see the "{{nth:nth}}" {thing} "{{name}}"')
        def should_see_the_nth(ctx, nth, name):
            retry(
                base_should_see_the,
                retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
                wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
            )(ctx, thing, name, index=nth)

        @step(f'I wait to see the "{{nth:nth}}" {thing} "{{name}}"')
        def wait_to_see_the_nth(ctx, nth, name):
            retry(base_should_see_the)(ctx, thing, name, index=nth)

        @step(
            f'I wait up to "{{seconds}}" seconds to see the "{{nth:nth}}" {thing} "{{name}}"'
        )
        def wait_up_to_seconds_to_see_the_nth(ctx, seconds, nth, name):
            seconds = float(seconds)
            retry(base_should_see_the, wait_up_to_s=seconds)(
                ctx, thing, name, index=nth
            )

    # should not see
    # undecorated def for reference below
    def base_should_not_see_the(ctx, thing, name, index=0):
        prefix = nth_to_ordinal(index)
        element = find_func(ctx, name, index=index)

        if element is not None:
            raise AssertionError(f'able to find the {prefix}{thing} "{name}"')
        logger.debug(f'Success: did not see {prefix}{thing} "{name}"')

    @step(f'I should immediately not see the {thing} "{{name}}"')
    def should_immediately_not_see_the(ctx, thing, name):
        base_should_not_see_the(ctx, thing, name)

    @step(f'I should not see the {thing} "{{name}}"')
    def should_not_see_the(ctx, name):
        retry(
            base_should_not_see_the,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, thing, name)

    @step(f'I wait to not see the {thing} "{{name}}"')
    def wait_to_not_see_the(ctx, name):
        retry(base_should_not_see_the)(ctx, thing, name)

    @step(
        f'I wait up to "{{seconds}}" seconds to not see the {thing} "{{name}}"'
    )
    def wait_up_to_seconds_to_not_see_the(ctx, seconds, name):
        milliseconds = float(seconds)
        retry(base_should_not_see_the, wait_up_to_s=milliseconds)(
            ctx, thing, name
        )

    if with_nth:

        @step(
            f'I should immediately not see the "{{nth:nth}}" {thing} "{{name}}"'
        )
        def should_immediately_not_see_the_nth(ctx, nth, name):
            base_should_not_see_the(ctx, thing, name, index=nth)

        @step(f'I should not see the "{{nth:nth}}" {thing} "{{name}}"')
        def should_not_see_the_nth(ctx, nth, name):
            retry(
                base_should_not_see_the,
                retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
                wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
            )(ctx, thing, name, index=nth)

        @step(f'I wait to not see the "{{nth:nth}}" {thing} "{{name}}"')
        def wait_to_not_see_the_nth(ctx, nth, name):
            retry(base_should_not_see_the)(ctx, thing, name, index=nth)

        @step(
            f'I wait up to "{{seconds}}" seconds to not see the "{{nth:nth}}" {thing} "{{name}}"'
        )
        def wait_up_to_seconds_to_not_see_the_nth(ctx, seconds, nth, name):
            seconds = float(seconds)
            retry(base_should_not_see_the, wait_up_to_s=seconds)(
                ctx, thing, name, index=nth
            )


def define_action_on_thing_with_name_steps(
    thing, action, find_func, action_func, with_nth=False
):
    """
    defines steps with with the following signatures:

      I {action} the {thing} "{name}"
      I wait to {action} the {thing} "{name}"
      I wait up to "{seconds}" seconds to {action} the {thing} "{name}"
      ...
      I {action} the {thing} "{name}" if it exists

      when with_nth=True we also define:

      I {action} the "{nth}" {thing} "{name}"
      I wait to {action} the "{nth}" {thing} "{name}"
      I wait up to "{seconds}" seconds to "{nth}" {action} the {thing} "{name}"
      ...
      I {action} the "{nth}" {thing} "{name}" if it exists

    parameters:
        thing(string):       name of the thing we're creating the steps for such
                             as button, dialog, etc.
        action(string):      the name of the action being performed, such as:
                             click, disable, etc.
        find_func(function): function that returns the desired element:

                             def find_func(ctx, name, index=):
                                '''
                                ctx(object):  behave context object
                                name(string): name of the thing to find
                                index(int):   when there are multiple elements
                                              with the same name and you've
                                              specified with_nth=True
                                '''
        action_func(function): function that performs the desired action:

                               def action_func(ctx, element):
                                  '''
                                  ctx(object):  behave context object
                                  element(object): the element found
                                  '''
        with_nth(bool):     when set to True we'll define the expanded set of
                            "nth" steps. default: False
    """

    # undecorated def for reference below
    def base_action_the(ctx, thing, name, index=0, must_exist=True):
        prefix = nth_to_ordinal(index)
        element = find_func(ctx, name, index=index)

        if element is None:
            if must_exist:
                raise AssertionError(
                    f'unable to find the {prefix}{thing} "{name}"'
                )
        else:
            action_func(ctx, element)
            logger.debug(
                f'Successfully executed {action} {prefix}{thing} "{name}"'
            )

    @step(f'I immediately {action} the {thing} "{{name}}"')
    def immediately_action_the(ctx, name):
        base_action_the(ctx, thing, name)

    @step(f'I {action} the {thing} "{{name}}"')
    def action_the(ctx, name):
        retry(
            base_action_the,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, thing, name)

    @step(f'I immediately {action} the {thing} "{{name}}" if it exists')
    def immediately_action_the_if_it_exists(ctx, name):
        base_action_the(ctx, thing, name, must_exist=False)

    @step(f'I {action} the {thing} "{{name}}" if it exists')
    def action_the_if_it_exists(ctx, name):
        retry(
            base_action_the,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, thing, name, must_exist=False)

    @step(f'I wait to {action} the {thing} "{{name}}"')
    def wait_to_action_the(ctx, name):
        retry(base_action_the)(ctx, thing, name)

    @step(f'I wait to {action} the {thing} "{{name}}" if it exists')
    def wait_to_action_the_if_it_exists(ctx, name):
        retry(base_action_the)(ctx, thing, name, must_exist=False)

    @step(
        f'I wait up to "{{seconds}}" seconds to {action} the {thing} "{{name}}"'
    )
    def wait_up_to_seconds_to_action_the(ctx, seconds, name):
        seconds = float(seconds)
        retry(base_action_the, wait_up_to_s=seconds)(ctx, thing, name)

    if with_nth:

        @step(f'I immediately {action} the "{{nth:nth}}" {thing} "{{name}}"')
        def immediately_action_the_nth(ctx, nth, name):
            base_action_the(ctx, thing, name, index=nth)

        @step(f'I {action} the "{{nth:nth}}" {thing} "{{name}}"')
        def action_the_nth(ctx, nth, name):
            retry(
                base_action_the,
                retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
                wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
            )(ctx, thing, name, index=nth)

        @step(
            f'I immediately {action} the "{{nth:nth}}" {thing} "{{name}}" if it exists'
        )
        def immediately_action_the_nth_if_it_exists(ctx, nth, name):
            base_action_the(ctx, thing, name, index=nth, must_exist=False)

        @step(f'I {action} the "{{nth:nth}}" {thing} "{{name}}" if it exists')
        def action_the_nth_if_it_exists(ctx, nth, name):
            retry(
                base_action_the,
                retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
                wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
            )(ctx, thing, name, index=nth, must_exist=False)

        @step(f'I wait to {action} the "{{nth:nth}}" {thing} "{{name}}"')
        def wait_to_action_the_nth(ctx, nth, name):
            retry(base_action_the)(ctx, thing, name, index=nth)

        @step(
            f'I wait up to "{{seconds}}" seconds to {action} the "{{nth:nth}}" {thing} "{{name}}"'
        )
        def wait_up_to_action_the_nth(ctx, seconds, nth, name):
            seconds = float(seconds)
            retry(base_action_the, wait_up_to_s=seconds)(
                ctx, thing, name, index=nth
            )


def define_ensure_state_on_thing_with_name_steps(
    thing, state, find_func, state_func, action_func, with_nth=False
):
    """
    defines steps with with the following signatures:

      I ensure the {thing} "{name}" is {state}
      I wait to ensure the {thing} "{name}" is {state}
      I wait up to "{seconds}" seconds to ensure the {thing} "{name}" is {state}

      when with_nth=True we also define:

      I ensure the "{nth}" {thing} "{name}" is {state}
      I wait to ensure the "{nth}" {thing} "{name}" is {state}
      I wait up to "{seconds}" seconds to ensure the "{nth}" {thing} "{name}" is {state}

    parameters:
        thing(string):       name of the thing we're creating the steps for such
                             as button, dialog, etc.
        state(stirng):       the name of the state we want to ensure, such as:
                             filled, checked, empty, not empty, etc.
        find_func(function): function that returns the desired element:

                             def find_func(ctx, name, index=):
                                '''
                                ctx(object):  behave context object
                                name(string): name of the thing to find
                                index(int):   when there are multiple elements
                                              with the same name and you've
                                              specified with_nth=True
                                '''
        state_func(function): function that returns True if the element is in
                              the desired state and False otherwise:

                               def state_func(element):
                                  '''
                                  element(object): the element found
                                  '''
        action_func(function): action that will set the desired state on the
                               element if state_func returns False:
                               def action_func(ctx, element):
                                  '''
                                  ctx(object):  behave context object
                                  element(object): the element found
                                  '''
        with_nth(bool):     when set to True we'll define the expanded set of
                            "nth" steps. default: False
    """

    # undecorated def for reference below
    def base_ensure_the(ctx, thing, name, index=0):
        prefix = nth_to_ordinal(index)
        element = find_func(ctx, name, index=index)

        if element is None:
            raise AssertionError(
                f'unable to find the {prefix}{thing} "{name}"'
            )

        if not state_func(element):
            action_func(ctx, element)

    @step(f'I immediately ensure the {thing} "{{name}}" is {state}')
    def immediately_ensure_the(ctx, name, state):
        base_ensure_the(ctx, thing, name)

    @step(f'I ensure the {thing} "{{name}}" is {state}')
    def ensure_the(ctx, name):
        retry(
            base_ensure_the,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, thing, name)

    @step(f'I wait to ensure the {thing} "{{name}}" is {state}')
    def wait_to_ensure_the(ctx, name):
        retry(base_ensure_the)(ctx, thing, name)

    @step(
        f'I wait up to "{{seconds}}" seconds to ensure the {thing} "{{name}}" is {state}'
    )
    def wait_up_to_seconds_to_ensure_the(ctx, seconds, name):
        seconds = float(seconds)
        retry(base_ensure_the, wait_up_to_s=seconds)(ctx, thing, name)

    if with_nth:

        @step(
            f'I immediately ensure the "{{nth:nth}}" {thing} "{{name}}" is {state}'
        )
        def immediately_ensure_the_nth(ctx, nth, name):
            base_ensure_the(ctx, thing, name, index=nth)

        @step(f'I ensure the "{{nth:nth}}" {thing} "{{name}}"')
        def ensure_the_nth(ctx, nth, name):
            retry(
                base_ensure_the,
                retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
                wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
            )(ctx, thing, name, index=nth)

        @step(
            f'I wait to ensure the "{{nth:nth}}" {thing} "{{name}}" is {state}'
        )
        def wait_to_ensure_the_nth(ctx, nth, name):
            retry(base_ensure_the)(ctx, thing, name, index=nth)

        @step(
            f'I wait up to "{{seconds}}" seconds to ensure the "{{nth:nth}}" {thing} "{{name}}" is {state}'
        )
        def wait_up_to_ensure_the_nth(ctx, seconds, nth, name):
            seconds = float(seconds)
            retry(base_ensure_the, wait_up_to_s=seconds)(
                ctx, thing, name, index=nth
            )


def define_thing_with_name_in_state_steps(
    thing, state, find_func, is_in_state_func, with_nth=False
):
    """
    defines steps with with the following signatures:

      I should immediately see the {thing} "{name}" is {state_name}
      I should see the {thing} "{name}" is {state_name}
      I wait to see the {thing} "{name}" is {state_name}
      I wait up to "{seconds}" seconds to see the {thing} "{name}" is {state_name}

      when with_nth=True we also define:

      I should immediately see the "{nth}" {thing} "{name}" is {state_name}
      I should see the "{nth}" {thing} "{name}" is {state_name}
      I wait to see the "{nth}" {thing} "{name}" is {state_name}
      I wait up to "{seconds}" seconds to see the "{nth}" {thing} "{name}" is {state_name}

    parameters:
        thing(string):             name of the thing we're creating the steps for such
                                   as button, dialog, etc.
        state(stirng):             the name of the state being verified, such as:
                                   selected, checked, disabled, etc.
        find_func(function):       function that returns the desired element:

                                   def find_func(ctx, name, index=):
                                      '''
                                      ctx(object):  behave context object
                                      name(string): name of the thing to find
                                      index(int):   when there are multiple elements
                                                    with the same name and you've
                                                    specified with_nth=True
                                      '''
        is_int_state_func(function): function that verifies the element is in
                                     the desired state:

                                     def is_in_state_func(element):
                                        '''
                                        element(object): the element found
                                        returns(bool): returns True if in
                                                       expected state
                                        '''
        with_nth(bool):     when set to True we'll define the expanded set of
                            "nth" steps. default: False

    """

    # undecorated def for reference below
    def base_should_see_the_in_state(ctx, thing, name, index=0):
        prefix = nth_to_ordinal(index)
        element = find_func(ctx, name, index=index)

        if element is None:
            raise AssertionError(
                f'unable to find the {prefix}{thing} "{name}"'
            )

        if not is_in_state_func(element):
            raise AssertionError(f'{thing} "{name}" is not {state}')
        logger.debug(f'{thing} {name} was in desired state "{state}"')

    @step(f'I should immediately see the {thing} "{{name}}" is {state}')
    def should_immediately_see_the_in_state(ctx, name, index=0):
        base_should_see_the_in_state(ctx, thing, name, index=0)

    @step(f'I should see the {thing} "{{name}}" is {state}')
    def should_see_the_in_state(ctx, name):
        retry(
            base_should_see_the_in_state,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, thing, name)

    @step(f'I wait to see the {thing} "{{name}}" is {state}')
    def wait_to_see_the_in_state(ctx, name):
        retry(base_should_see_the_in_state)(ctx, thing, name)

    @step(
        f'I wait up to "{{seconds}}" seconds to see the {thing} "{{name}}" is {state}'
    )
    def wait_up_to_seconds_to_see_the_in_state(ctx, seconds, name):
        seconds = float(seconds)
        retry(base_should_see_the_in_state, wait_up_to_s=seconds)(
            ctx, thing, name
        )

    if with_nth:

        @step(
            f'I should immediately see the "{{nth:nth}}" {thing} "{{name}}" is {state}'
        )
        def base_should_see_the_nth_in_state(ctx, nth, name):
            base_should_see_the_in_state(ctx, thing, name, index=nth)

        @step(f'I should see the "{{nth:nth}}" {thing} "{{name}}" is {state}')
        def should_see_the_nth_in_state(ctx, nth, name):
            retry(
                base_should_see_the_in_state,
                retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
                wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
            )(ctx, thing, name, index=nth)

        @step(f'I wait to see the "{{nth:nth}}" {thing} "{{name}}" is {state}')
        def wait_to_see_the_nth_in_state(ctx, nth, name):
            retry(base_should_see_the_in_state)(ctx, thing, name, index=nth)

        @step(
            f'I wait up to "{{seconds}}" seconds to see the "{{nth:nth}}" {thing} "{{name}}" is {state}'
        )
        def wait_up_to_seconds_to_see_the_nth_in_state(
            ctx, seconds, nth, name
        ):
            seconds = float(seconds)
            retry(base_should_see_the_in_state, wait_up_to_s=seconds)(
                ctx, thing, name, index=nth
            )


def define_run_steps_if_I_can_see_element_with_name_steps(thing, find_func):
    """
    defines steps with with the following signatures:

      I run the following steps if I can immediately see the {thing} "{name}"
      I run the following steps if I can see the {thing} "{name}"

      I run the following steps if I can immediately not see the {thing} "{name}"
      I run the following steps if I can not see the {thing} "{name}"

    parameters:
        thing(string):       name of the thing we're creating the steps for such
                             as button, dialog, etc.
        find_func(function): function that returns the element that matches the
                             name provided and is visible

                             def find_func(ctx, name):
                                '''
                                ctx(object):  behave context object
                                name(string): name of the thing to find
                                '''
    """

    # undecorated def for reference below
    def base_run_if_visibile(ctx, name):
        element = find_func(ctx, name)

        if element is not None:
            run_steps(ctx, ctx.text)

    @step(
        f'I run the following steps if I can immediately see the {thing} "{{name}}"'
    )
    def run_if_immediately_visibile(ctx, name):
        base_run_if_visibile(ctx, name)

    @step(f'I run the following steps if I can see the {thing} "{{name}}"')
    def run_if_visibile(ctx, name):
        retry(
            base_run_if_visibile,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, name)

    def base_run_if_not_visibile(ctx, name):
        element = find_func(ctx, name)

        if element is None:
            run_steps(ctx, ctx.text)

    @step(
        f'I immediately run the following steps if I can not see the {thing} "{{name}}"'
    )
    def immediately_run_if_not_visibile(ctx, name):
        base_run_if_not_visibile(ctx, name)

    @step(f'I run the following steps if I can not see the {thing} "{{name}}"')
    def run_if_not_visibile(ctx, name):
        retry(
            base_run_if_not_visibile,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, name)


def define_two_thing_interaction_steps(
    action: str,
    action_func,
    thing_1,
    thing_1_find_func,
    preposition: str,
    thing_2,
    thing_2_find_func,
    with_nth=False,
):
    """
    defines steps with with the following signatures:
      I {action} the {thing_1} "{name_1}" {preposition} the {thing_2} "{name_2}"
      I wait to {action} the {thing_1} "{name_1}" {preposition} the {thing_2} "{name_2}"
      I wait up to "{seconds}" seconds to {action} the {thing_1} "{name_1}" {preposition} the {thing_2} "{name_2}"
      ...
      I {action} the {thing_1} "{name_1}" {preposition} the {thing_2} "{name_2}" if they both exist


      when with_nth=True we also define:

      I {action} the "{nth_1}" {thing_1} "{name_1}" {preposition} the "{nth_2}" {thing_2} "{name_2}"
      I wait to {action} the "{nth_1}" {thing_1} "{name_1}" {preposition} the "{nth_2}" {thing_2} "{name_2}"
      I wait up to "{seconds}" seconds to {action} the "{nth_1}" {thing_1} "{name_1}" {preposition} the "{nth_2}" {thing_2} "{name_2}"
      ...
      I {action} the "{nth_1}" {thing_1} "{name_1}" {preposition} the "{nth_2}" {thing_2} "{name_2}" if they both exist

    parameters:
        action(string):      the name of the action being performed, such as:
                             click, disable, etc.
        action_func(function):      function that performs the desired action:

                                    def action_func(ctx, element, ):
                                        '''
                                        ctx(object):  behave context object
                                        element(object): the element found
                                        '''
        thing_1(string):     name of the thing we're creating the steps for such
                             as button, dialog, etc.
        thing_1_find_func(function): function that returns the desired element:

                                     def thing_1_find_func(ctx, name_1, index_1=):
                                        '''
                                        ctx(object):   behave context object
                                        name_1(string):name of the thing to find
                                        index_1(int):    when there are multiple elements
                                                    with the same name and you've
                                                    specified with_nth=True
                                        '''
        preposition(string):    preposition to help with readability as there are
                                many different prepositions that would be valid for
                                a desired action
        thing_2(string):     name of the thing that is being interacted with
                             from the defined action
        thing_2_find_func(function): function that returns the interacted element:

                                     def thing_2_find_func(ctx, name_2, index_2=):
                                        '''
                                        ctx(object):    behave context object
                                        name_2(string): name of the thing to find
                                        index_1(int):     when there are multiple elements
                                                    with the same name and you've
                                                    specified with_nth=True
                                        '''
        with_nth(bool):      when set to True we'll define the expanded set of
                             "nth" steps. default: False
    """

    # undecorated def for reference below
    def base_action_the(
        ctx,
        thing_1,
        name_1,
        thing_2,
        name_2,
        index_1=0,
        index_2=0,
    ):
        prefix_1 = nth_to_ordinal(index_1)
        prefix_2 = nth_to_ordinal(index_2)

        element_1 = thing_1_find_func(ctx, name_1, index_1)
        element_2 = thing_2_find_func(ctx, name_2, index_2)

        if element_1 is None or element_2 is None:
            error_message = []
            if element_1 is None:
                error_message.append(
                    f'Unable to find the {prefix_1}{thing_1} "{name_1}"'
                )
            if element_2 is None:
                error_message.append(
                    f'Unable to find the {prefix_2}{thing_2} "{name_2}"'
                )

            raise AssertionError(", ".join(error_message))

        else:
            action_func(ctx, element_1, element_2)
            logger.debug(
                f'Successfully executed {action} {prefix_1}{thing_1} "{name_1}" {preposition} {prefix_2}{thing_2} "{name_2}"'
            )

    @step(
        f'I immediately {action} the {thing_1} "{{name_1}}" {preposition} the {thing_2} "{{name_2}}"'
    )
    def immediately_action_the(ctx, name_1, name_2):
        base_action_the(ctx, thing_1, name_1, thing_2, name_2)

    @step(
        f'I {action} the {thing_1} "{{name_1}}" {preposition} the {thing_2} "{{name_2}}"'
    )
    def action_the(ctx, name_1, name_2):
        retry(
            base_action_the,
            retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
            wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
        )(ctx, thing_1, name_1, thing_2, name_2)

    @step(
        f'I wait to {action} the {thing_1} "{{name_1}}" {preposition} the {thing_2} "{{name_2}}"'
    )
    def wait_to_action_the(ctx, name_1, name_2):
        retry(base_action_the)(ctx, thing_1, name_1, thing_2, name_2)

    @step(
        f'I wait up to "{{seconds}}" seconds to {action} the {thing_1} "{{name_1}}" {preposition} the {thing_2} "{{name_2}}"'
    )
    def wait_up_to_seconds_to_action_the(ctx, seconds, name_1, name_2):
        seconds = float(seconds)
        retry(base_action_the, wait_up_to_s=seconds)(
            ctx, thing_1, name_1, thing_2, name_2
        )

    if with_nth:

        @step(
            f'I immediately {action} the "{{nth_1:nth}}" {thing_1} "{{name_1}}" {preposition} the "{{nth_2:nth}}" {thing_2} "{{name_2}}"'
        )
        def immediately_action_the_nth_i_nth(
            ctx, nth_1, name_1, nth_2, name_2
        ):
            base_action_the(
                ctx,
                thing_1,
                name_1,
                thing_2,
                name_2,
                index_1=nth_1,
                index_2=nth_2,
            )

        @step(
            f'I {action} the "{{nth_1:nth}}" {thing_1} "{{name_1}}" {preposition} the "{{nth_2:nth}}" {thing_2} "{{name_2}}"'
        )
        def action_the_nth_i_nth(ctx, nth_1, name_1, nth_2, name_2):
            retry(
                base_action_the,
                retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
                wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
            )(
                ctx,
                thing_1,
                name_1,
                thing_2,
                name_2,
                index_1=nth_1,
                index_2=nth_2,
            )

        @step(
            f'I wait to {action} the "{{nth_1:nth}}" {thing_1} "{{name_1}}" {preposition} the "{{nth_2:nth}}" {thing_2} "{{name_2}}"'
        )
        def wait_to_action_the_nth_ith(ctx, nth_1, name_1, nth_2, name_2):
            retry(base_action_the)(
                ctx,
                thing_1,
                name_1,
                thing_2,
                name_2,
                index_1=nth_1,
                index_2=nth_2,
            )

        @step(
            f'I wait up to "{{seconds}}" seconds to {action} the "{{nth_1:nth}}" {thing_1} "{{name_1}}" {preposition} the "{{nth_2:nth}}" {thing_2} "{{name_2}}"'
        )
        def wait_up_to_action_the_nth_i_nth(
            ctx, seconds, nth_1, name_1, nth_2, name_2
        ):
            seconds = float(seconds)
            retry(base_action_the, wait_up_to_s=seconds)(
                ctx,
                thing_1,
                name_1,
                thing_2,
                name_2,
                index_1=nth_1,
                index_2=nth_2,
            )



import operator
import pkgutil
import re
from io import StringIO

from selenium.webdriver.common.by import By

from cucu import (
    config,
    format_gherkin_table,
    fuzzy,
    helpers,
    logger,
    retry,
    step,
)
from cucu.browser.frames import run_in_all_frames


def find_tables(ctx):
    """
    find all the tables currently present on the page

    parameters:
      ctx(object): behave context object used to share data between steps

    returns:
      an array of arrays containing the HTML tables currently displayed
    """
    ctx.check_browser_initialized()
    tables_lib = pkgutil.get_data("cucu", "steps/tables.js")
    tables_lib = tables_lib.decode("utf8")

    def search_for_tables():
        ctx.browser.execute(tables_lib)
        return ctx.browser.execute("return findAllTables();")

    return run_in_all_frames(ctx.browser, search_for_tables)


def behave_table_to_array(table):
    """
    given a behave.model.Table object convert it to an array of rows

    parameters:
        table(behave.model.Table): the behave table to convert into an array

    returns:
        array of rows representing the behave table provided.
    """
    result = [table.headings]

    for row in table.rows:
        values = []
        for value in row:
            values.append(value)
        result.append(values)

    return result


def check_table_equals_table(table, expected_table):
    """
    check that table is equal to expected_table
    """
    return table == expected_table


def check_table_matches_table(table, expected_table):
    """
    check if table matches the regex patterns in expected table
    """

    table_matched = bool(len(table) == len(expected_table))
    if table_matched:
        for expected_row, row in zip(expected_table, table):
            for expected_value, value in zip(expected_row, row):
                if not re.match(expected_value, value):
                    table_matched = False
                    break

    return table_matched


def check_table_contains_table(table, expected_table):
    """
    check that table contains the rows in expected_table
    """
    return all(row in table for row in expected_table)


def check_table_contains_matching_rows_in_table(table, expected_table):
    """
    check that table contains the matching rows in expected_table
    """
    for expected_row in expected_table:
        found_row = False
        for row in table:
            good_cols = len(expected_row) == len(row)

            for expected_value, value in zip(expected_row, row):
                if not re.match(expected_value, value):
                    good_cols = False

            if good_cols:
                found_row = True
                break

        if not found_row:
            return False

    return True


def find_header_row_index(table, expected_headers, max_scan_rows=2):
    """
    Find first header row (within first 2 rows) where each expected header regex
    matches at least one actual header cell.
    """
    expected_patterns = [(h or "").strip(" ") for h in expected_headers]
    scan_rows = min(len(table), max_scan_rows)

    for idx in range(scan_rows):
        actual_headers = [(v or "").strip(" ") for v in table[idx]]

        all_headers_present = True
        for pattern in expected_patterns:
            if not any(re.match(pattern, actual) for actual in actual_headers):
                all_headers_present = False
                break

        if all_headers_present:
            return idx

    return None


def build_expected_header_to_indices(header_row, expected_headers):
    """
    For each expected header regex, return all matching column indices
    from the actual header row (supports duplicate headers / colspan).
    """
    actual_headers = [(h or "").strip(" ") for h in header_row]
    expected_patterns = [(h or "").strip(" ") for h in expected_headers]

    expected_to_indices = []
    for pattern in expected_patterns:
        indices = [
            idx
            for idx, actual in enumerate(actual_headers)
            if re.match(pattern, actual)
        ]
        if not indices:
            return None
        expected_to_indices.append(indices)

    return expected_to_indices


def contains_rows_by_named_columns(table, expected_table, match_value):
    """
    expected_table format (from behave_table_to_array(ctx.table)):
      expected_table[0] = expected header regexes (only columns you care about)
      expected_table[1:] = expected cell values/patterns for those headers

    For each expected row, find at least one actual data row that matches all
    specified columns. Extra columns and column order in the UI are ignored.
    """
    if not expected_table or len(expected_table) < 2:
        return False

    expected_headers = expected_table[0]
    expected_rows = expected_table[1:]

    header_row_idx = find_header_row_index(
        table, expected_headers, max_scan_rows=2
    )
    if header_row_idx is None:
        return False

    expected_to_indices = build_expected_header_to_indices(
        table[header_row_idx], expected_headers
    )
    if expected_to_indices is None:
        return False

    actual_data_rows = table[header_row_idx + 1 :]

    for expected_row in expected_rows:
        if len(expected_row) != len(expected_headers):
            return False

        matched_any_actual_row = False

        for actual_row in actual_data_rows:
            row_ok = True

            for expected_col_idx, expected_value in enumerate(expected_row):
                candidate_indices = expected_to_indices[expected_col_idx]

                matched_this_expected_column = False
                for col_idx in candidate_indices:
                    if col_idx < len(actual_row) and match_value(
                        expected_value, actual_row[col_idx]
                    ):
                        matched_this_expected_column = True
                        break

                if not matched_this_expected_column:
                    row_ok = False
                    break

            if row_ok:
                matched_any_actual_row = True
                break

        if not matched_any_actual_row:
            return False

    return True


def check_table_contains_rows_matching_only_these_columns(
    table, expected_table
):
    """Header and cell values matched via regex."""
    return contains_rows_by_named_columns(
        table,
        expected_table,
        lambda expected, actual: re.match(expected, actual),
    )


def check_table_contains_rows_with_only_these_columns(table, expected_table):
    """Header matched via regex; cell values matched by exact equality."""
    return contains_rows_by_named_columns(
        table, expected_table, lambda expected, actual: expected == actual
    )


def report_found_table(expected_table, found_table):
    logger.debug(
        f"found desired table\nexpected:\n{format_gherkin_table(expected_table, [], '  ')}\n\n"
        f"found:\n{format_gherkin_table(found_table, [], '  ')}\n"
    )


def report_unable_to_find_table(expected_table, found_tables):
    stream = StringIO()
    stream.write("\n")
    for index, table in enumerate(found_tables):
        print_index = helpers.nth_to_ordinal(index) or '"1st" '
        stream.write(
            f"{print_index}table:\n{format_gherkin_table(table, [], '  ')}\n"
        )

    stream.seek(0)
    raise AssertionError(
        f"unable to find desired table\nexpected:\n{format_gherkin_table(expected_table, [], '  ')}\n\nfound:{stream.read()}"
    )


def report_found_undesired_table(unexpected_tables, found_tables):
    stream = StringIO()
    stream.write("\n")
    for index, table in enumerate(found_tables):
        print_index = helpers.nth_to_ordinal(index) or '"1st" '
        stream.write(
            f"{print_index}table:\n{format_gherkin_table(table, [], '  ')}\n"
        )

    stream.seek(0)
    error_message = ""
    for table in unexpected_tables:
        error_message += (
            f"found undesired table\n\nundesired table:\n{format_gherkin_table(table, [], '  ')}\n\n"
            f"all tables found:{stream.read()}\n"
        )
    raise AssertionError(error_message)


def find_table(ctx, assert_func, nth=None):
    """
    validate we can find the table passed in the ctx object and assert it
    matches anyone of the tables on the current web page. If `nth` is set to
    something then we only check against the nth table of the available tables.

    parameters:
        ctx(object): behave context object
        assert_func(function): function used to assert two tables "match"
        nth(int): when set to an int specifies the exact table within the list
                  of available tables to match against.

    returns:
        array of arrays representing the matching HTML table

    raises:
        AssertionError when the desired table was not found
    """
    expected = behave_table_to_array(ctx.table)
    found_tables = find_tables(ctx)

    if nth is not None:
        if assert_func(found_tables[nth], expected):
            report_found_table(expected, found_tables[nth])
            return found_tables[nth]

    else:
        for table in found_tables:
            if assert_func(table, expected):
                report_found_table(expected, table)
                return table

    report_unable_to_find_table(expected, found_tables)


def do_not_find_table(ctx, assert_func, nth=None):
    """
    validate we can not find the table passed in the ctx object and assert it
    matches anyone of the tables on the current web page. If `nth` is set to
    something then we only check against the nth table of the available tables.

    paramters:
        ctx(object): behave context object
        assert_func(function): function used to assert two tables "match"
        nth(int): when set to an int specifies the exact table within the list
                  of available tables to match against.

    raises:
        AssertionError when the desired table was not found
    """
    expected = behave_table_to_array(ctx.table)
    tables = find_tables(ctx)
    matching_tables = []

    if nth is not None:
        if assert_func(tables[nth], expected):
            matching_tables.append(tables[nth])

    else:
        for table in tables:
            if assert_func(table, expected):
                matching_tables.append(table)

    # If none of the tables match the pattern, then return
    if len(matching_tables) == 0:
        return

    report_found_undesired_table(matching_tables, tables)


for thing, check_func in {
    "is": check_table_equals_table,
    "matches": check_table_matches_table,
    "contains": check_table_contains_table,
    "contains rows matching": check_table_contains_matching_rows_in_table,
}.items():

    @step(f"I should see a table that {thing} the following")
    def should_see_the_table(ctx, check_func=check_func):
        find_table(ctx, check_func)

    @step(f"I should not see a table that {thing} the following")
    def should_not_see_the_table(ctx, check_func=check_func):
        do_not_find_table(ctx, check_func)

    @step(f"I wait to see a table that {thing} the following")
    def wait_to_see_the_table(ctx, check_func=check_func):
        retry(find_table)(ctx, check_func)

    @step(f"I wait to not see a table that {thing} the following")
    def wait_to_not_see_the_table(ctx, check_func=check_func):
        retry(do_not_find_table)(ctx, check_func)

    @step(
        f'I wait up to "{{seconds}}" seconds to see a table that {thing} the following'
    )
    def wait_up_to_seconds_to_see_the_table(
        ctx, seconds, check_func=check_func
    ):
        seconds = float(seconds)
        retry(find_table, wait_up_to_s=seconds)(ctx, check_func)

    @step(
        f'I wait up to "{{seconds}}" seconds to not see a table that {thing} the following'
    )
    def wait_up_to_seconds_to_not_see_the_table(
        ctx, seconds, check_func=check_func
    ):
        seconds = float(seconds)
        retry(do_not_find_table, wait_up_to_s=seconds)(ctx, check_func)

    @step(f'I should see the "{{nth}}" table {thing} the following')
    def should_see_the_nth_table(ctx, nth, check_func=check_func):
        find_table(ctx, check_func, nth=nth)

    @step(f'I wait to see the "{{nth}}" table {thing} the following')
    def wait_to_see_the_nth_table(ctx, nth, check_func=check_func):
        retry(find_table)(ctx, check_func, nth=nth)

    @step(
        f'I wait up to "{{seconds}}" seconds to see the "{{nth}}" table {thing} the following'
    )
    def wait_up_to_seconds_to_see_the_nth_table(
        ctx, seconds, nth, check_func=check_func
    ):
        seconds = float(seconds)
        retry(find_table, wait_up_to_s=seconds)(ctx, check_func, nth=nth)


@step("I should see a table with rows matching these columns and values")
def should_see_table_rows_matching_only_these_columns(ctx):
    find_table(ctx, check_table_contains_rows_matching_only_these_columns)


@step(
    "I should not see a table that contains rows matching these columns and values"
)
def should_not_see_table_rows_matching_only_these_columns(ctx):
    do_not_find_table(
        ctx, check_table_contains_rows_matching_only_these_columns
    )


@step(
    "I wait to see a table that contains rows matching these columns and values"
)
def wait_to_see_table_rows_matching_only_these_columns(ctx):
    retry(find_table)(
        ctx, check_table_contains_rows_matching_only_these_columns
    )


@step(
    'I wait up to "{seconds}" seconds to see a table that contains rows matching these columns and values'
)
def wait_up_to_seconds_to_see_table_rows_matching_only_these_columns(
    ctx, seconds
):
    retry(find_table, wait_up_to_s=float(seconds))(
        ctx, check_table_contains_rows_matching_only_these_columns
    )


@step(
    'I should see the "{nth}" table containing rows matching these columns and values'
)
def should_see_nth_table_rows_matching_only_these_columns(ctx, nth):
    find_table(
        ctx, check_table_contains_rows_matching_only_these_columns, nth=nth
    )


@step(
    'I wait to see the "{nth}" table containing rows matching these columns and values'
)
def wait_to_see_nth_table_rows_matching_only_these_columns(ctx, nth):
    retry(find_table)(
        ctx, check_table_contains_rows_matching_only_these_columns, nth=nth
    )


def find_table_header(ctx, name, index=0):
    """
    find a table header with the provided name
    """
    ctx.check_browser_initialized()
    return fuzzy.find(ctx.browser, name, ["th"], index=index)


def click_table_header(ctx, header):
    """
    internal method used to simply click a table header element
    """
    ctx.check_browser_initialized()
    ctx.browser.click(header)


helpers.define_action_on_thing_with_name_steps(
    "table header",
    "click",
    find_table_header,
    click_table_header,
    with_nth=True,
)


def get_table_cell_value(ctx, table, row, column, variable_name):
    tables = find_tables(ctx)

    try:
        cell_value = tables[table][row][column]
    except IndexError:
        raise AssertionError(
            f"Cannot find table:{table + 1},row:{row + 1},column:{column + 1}. Please check your table data."
        )
    config.CONFIG[variable_name] = cell_value


@step(
    'I save "{table:nth}" table, "{row:nth}" row, "{column:nth}" column value to a variable "{variable_name}"'
)
def step_get_table_cell_value(ctx, table, row, column, variable_name):
    get_table_cell_value(ctx, table, row, column, variable_name)


@step(
    'I wait to save "{table:nth}" table, "{row:nth}" row, "{column:nth}" column value to a variable "{variable_name}"'
)
def wait_to_get_table_cell_value(ctx, table, row, column, variable_name):
    retry(get_table_cell_value)(ctx, table, row, column, variable_name)


def find_table_element(ctx, nth=1):
    """
    Return the nth table as a WebElement

    parameters:
      ctx(object): behave context object used to share data between steps
      nth(int): specifies the exact table within the list of available tables to match against.
                Defaults to 1st table.

    returns:
      A selenium WebElement associated with the table that was specified
    """
    ctx.check_browser_initialized()

    try:
        return ctx.browser.css_find_elements("table", missing_ok=True)[nth]
    except IndexError:
        raise AssertionError(
            f"Cannot find table:{nth + 1}. Please check your table data."
        )


def click_table_cell(ctx, row, column, table):
    """
    Clicks the cell corresponding to the given row and column

    parameters:
      ctx(object): behave context object used to share data between steps
      row(int): the row of the table to click
      column(int): the column of the table to click
      table(int): specifies the exact table within the list of available tables to match against.
    """
    table_element = find_table_element(ctx, table)

    try:
        row = table_element.find_elements(By.CSS_SELECTOR, "tbody tr")[row]
        cell = row.find_elements(By.CSS_SELECTOR, "td")[column]
    except IndexError:
        raise AssertionError(
            f"Cannot find table:{table + 1},row:{row + 1},column:{column + 1}. Please check your table data."
        )
    ctx.browser.click(cell)


@step('I wait to click the "{row:nth}" row in the "{table:nth}" table')
def wait_click_table_row(ctx, row, table):
    """
    Add 1 to the row number if the table has a header row.

    Note: Firefox is unable to click directly on a row <tr> if it has child columns <td>.
    In order to workaround this, the step just clicks the first column <td> of the row <tr>.
    Bug: https://bugzilla.mozilla.org/show_bug.cgi?id=1448825
    """
    retry(click_table_cell)(ctx, row, 1, table)


@step(
    'I wait to click the cell corresponding to the "{row:nth}" row and "{column:nth}" column in the "{table:nth}" table'
)
def wait_click_table_cell(ctx, row, column, table):
    """
    Add 1 to the row number if the table has a header row.
    """
    retry(click_table_cell)(ctx, row, column, table)


@step(
    'I wait to click the "{column:nth}" column within a row that contains the text "{match_text}" in the "{table:nth}" table'
)
def wait_click_table_cell_matching_text(ctx, column, match_text, table):
    def click_table_cell_matching_text(ctx, column, match_text, table):
        table_element = find_table_element(ctx, table)

        try:
            row = table_element.find_elements(
                By.XPATH, f'//td[.="{match_text}"]/parent::tr'
            )
            if len(row) > 1:
                logger.warning(
                    f'Found {len(row)} rows with matching text "{match_text}", using the first row.'
                )
            cell = row[0].find_elements(By.CSS_SELECTOR, "td")[column]
        except IndexError:
            raise AssertionError(
                f"Cannot find table:{table + 1},column:{column + 1},text:{match_text}. Please check your table data."
            )

        ctx.browser.click(cell)

    retry(click_table_cell_matching_text)(ctx, column, match_text, table)


def count_rows_in_table_element(table_element):
    """
    Count the amount of rows the provided table element has.

    Rows are defined as tr elements.
    Note that this is not a behave table nor an array of rows, unlike the result of func:`find_table`.

    parameters:
        table(WebElement): the table to convert find trs from
    """
    table_rows = table_element.find_elements(By.CSS_SELECTOR, "tr")
    return len(table_rows)


def find_nth_table_and_validate_row_count(
    ctx, table, thing, check_func, row_count
):
    """
    Find the nth table by index and validate its row count using the provided check function

    parameters:
        ctx(object): behave context object used to share data between steps
        table(int): the index of the table to check (0-based)
        thing(str): description of the comparison operator (e.g., "more than", "equals")
        check_func(function): operator function to perform the comparison (e.g., operator.gt, operator.eq)
        row_count(str): expected row count as a string

    raises:
        AssertionError: when the table row count does not meet the specified criteria
    """
    table_element = find_table_element(ctx, table)
    # table_element is a WebElement when search is done using find_table_element
    table_rows = count_rows_in_table_element(table_element)
    if not check_func(table_rows, int(row_count)):
        raise AssertionError(
            f"Expected {thing} {row_count} rows in table {table + 1}, but found {table_rows} instead."
        )


def find_table_matching_rows_and_validate_row_count(
    ctx, thing, check_func, row_count
):
    """
    Find a table containing matching rows and validate its row count using the provided check function

    parameters:
        ctx(object): behave context object used to share data between steps
        thing(str): description of the comparison operator (e.g., "more than", "equals")
        check_func(function): operator function to perform the comparison (e.g., operator.gt, operator.eq)
        row_count(str): expected row count as a string

    raises:
        AssertionError: when the table row count does not meet the specified criteria
    """
    table_element = find_table(
        ctx, check_table_contains_matching_rows_in_table
    )
    # table_element is an array of rows when search is done using find_table
    table_rows = len(table_element)
    if not check_func(table_rows, int(row_count)):
        raise AssertionError(
            f"Expected {thing} {row_count} rows in the table contaning matching entries, but found {table_rows} instead."
        )


@step('I wait to see there are "{row_count}" rows in the "{table:nth}" table')
def wait_table_row_count(ctx, row_count, table):
    """
    Add 1 to the row number if the table has a header row.
    """

    def find_table_row_count(ctx, row_count, table):
        table_element = find_table_element(ctx, table)
        table_rows = count_rows_in_table_element(table_element)

        if not int(row_count) == table_rows:
            raise AssertionError(
                f"Unable to find {row_count} rows in table {table + 1}. Please check your table data."
            )

    retry(find_table_row_count)(ctx, row_count, table)


for thing, check_func in {
    "more than": operator.gt,
    "at least": operator.ge,
    "exactly": operator.eq,
    "at most": operator.le,
    "less than": operator.lt,
    "not equal to": operator.ne,
}.items():

    @step(
        f'I wait to see there are {thing} "{{row_count}}" rows in the "{{table:nth}}" table'
    )
    def should_see_the_table_with_row_count(
        ctx, row_count, table, thing=thing, check_func=check_func
    ):
        """
        Add 1 to the row number if the table has a header row.
        """
        retry(find_nth_table_and_validate_row_count)(
            ctx, table, thing, check_func, row_count
        )

    @step(
        f'I wait to see that the table containing these rows has {thing} "{{row_count}}" rows'
    )
    def should_see_the_table_containing_rows_with_row_count(
        ctx, row_count, thing=thing, check_func=check_func
    ):
        """
        Add 1 to the row count number if the table has a header row.
        """
        retry(find_table_matching_rows_and_validate_row_count)(
            ctx, thing, check_func, row_count
        )



import contextlib
import json
import os
import socket
import sys
from datetime import datetime
from pathlib import Path

from cucu import (
    behave_tweaks,
    init_global_hook_variables,
    register_before_retry_hook,
)
from cucu.browser import selenium
from cucu.config import CONFIG
from cucu.db import create_database_file, record_cucu_run
from cucu.page_checks import init_page_checks
from cucu.utils import behave_filepath_to_cucu_logpath, get_feature_name


def behave_init(filepath="features"):
    """
    behave internal init method used to load the various parts of set of
    feature files and supporting code without executing any of it.

    parameters:
        filepath(string): the file system path of the features directory to load
    """
    behave_tweaks.behave_main(
        ["--dry-run", "--format=null", "--no-summary", filepath]
    )


def behave(
    filepath,
    color_output,
    dry_run,
    env,
    fail_fast,
    headless,
    name,
    debug_on_failure,
    junit,
    results,
    secrets,
    show_skips,
    tags,
    verbose,
    redirect_output=False,
    skip_init_global_hook_variables=False,
):
    # load all them configs
    CONFIG.load_cucurc_files(filepath)

    if CONFIG["CUCU_SELENIUM_REMOTE_URL"] is None:
        selenium.init()

    # general socket timeout instead of letting the framework ever get stuck on a
    # socket connect/read call
    timeout = float(CONFIG["CUCU_SOCKET_DEFAULT_TIMEOUT_S"])
    socket.setdefaulttimeout(timeout)

    if not skip_init_global_hook_variables:
        init_global_hook_variables()

    init_page_checks()

    os.environ["CUCU_COLOR_OUTPUT"] = str(color_output).lower()

    if headless:
        os.environ["CUCU_BROWSER_HEADLESS"] = "True"

    for variable in list(env):
        key, value = variable.split("=")
        os.environ[key] = value

    if debug_on_failure:
        os.environ["CUCU_DEBUG_ON_FAILURE"] = "true"

    os.environ["CUCU_RESULTS_DIR"] = str(results)
    os.environ["CUCU_JUNIT_DIR"] = str(junit)

    if secrets:
        os.environ["CUCU_SECRETS"] = secrets

    args = [
        # don't run disabled tests
        "--tags",
        "~@disabled",
        # always print the skipped steps and scenarios
        "--show-skipped",
    ]

    if verbose:
        args.append("--verbose")

    run_json_filename = "run.json"
    if redirect_output:
        feature_name = get_feature_name(filepath)
        run_json_filename = f"{feature_name}-run.json"

    if dry_run:
        args += [
            "--dry-run",
            # console formatter
            "--format=cucu.formatter.cucu:CucuFormatter",
            # run.db formatter
            "--format=cucu.formatter.rundb:RundbFormatter",
        ]

    else:
        args += [
            "--no-capture",
            "--no-capture-stderr",
            "--no-logcapture",
            # generate a JSON file containing the exact details of the whole run
            "--format=cucu.formatter.json:CucuJSONFormatter",
            f"--outfile={results / run_json_filename}",
            # console formatter
            "--format=cucu.formatter.cucu:CucuFormatter",
            f"--logging-level={os.environ['CUCU_LOGGING_LEVEL'].upper()}",
            # disable behave's junit output in favor of our own formatter
            "--no-junit",
            "--format=cucu.formatter.junit:CucuJUnitFormatter",
            # run.db formatter
            "--format=cucu.formatter.rundb:RundbFormatter",
        ]

    for tag in tags:
        args.append("--tags")
        args.append(tag)

    if name is not None:
        args += ["--name", name]

    if fail_fast:
        args.append("--stop")

    if not show_skips:
        args.append("--no-skipped")

    args.append(filepath)
    os.environ["BEHAVE_FILEPATH"] = CONFIG["BEHAVE_FILEPATH"] = str(filepath)

    result = 0
    try:
        if redirect_output:
            cucu_log_path = behave_filepath_to_cucu_logpath(filepath, results)

            CONFIG["__CUCU_PARENT_STDOUT"] = sys.stdout

            def retry_progress(ctx):
                CONFIG["__CUCU_PARENT_STDOUT"].write(".")
                CONFIG["__CUCU_PARENT_STDOUT"].flush()

            # this allows steps that are stuck in a retry to loop to still
            # provide progress feedback on screen
            register_before_retry_hook(retry_progress)

            with cucu_log_path.open("w", encoding="utf8") as output:
                with contextlib.redirect_stderr(output):
                    with contextlib.redirect_stdout(output):
                        # intercept the stdout/stderr so we can do things such
                        # as hiding secrets in logs
                        behave_tweaks.init_outputs(sys.stdout, sys.stderr)
                        result = behave_tweaks.behave_main(args)
        else:
            # intercept the stdout/stderr so we can do things such
            # as hiding secrets in logs
            behave_tweaks.init_outputs(sys.stdout, sys.stderr)
            result = behave_tweaks.behave_main(args)
    except:
        result = -1
        raise

    return result


def create_run(results_path: Path, filepath: Path):
    run_json_filepath = results_path / "run_details.json"

    if run_json_filepath.exists():
        return

    env_values = (
        dict(os.environ)
        if CONFIG["CUCU_RECORD_ENV_VARS"]
        else "To enable use the --record-env-vars flag"
    )

    run_details = {
        "cucu_run_id": CONFIG["CUCU_RUN_ID"],
        "filepath": str(filepath),
        "full_arguments": sys.argv,
        "env": env_values,
        "date": datetime.now().isoformat(),
    }

    run_json_filepath.write_text(
        json.dumps(run_details, indent=2, sort_keys=True), encoding="utf8"
    )
    CONFIG["RUN_DB_PATH"] = run_db_path = results_path / "run.db"
    if not run_db_path.exists():
        create_database_file(run_db_path)

    record_cucu_run()



from cucu import run_steps, step


@step(
    'I login in with the username "{username}" and password as var "{password_var}"'
)
def step_with_substeps_that_log(ctx, username, password_var):
    """This is an example of a step that uses substeps.

    You probably want write a login step anyways"""
    run_steps(
        ctx,
        f"""
        When I write "{username}" into the input "username"
         And I write "{{{password_var}}}" into the input "password"
         And I click the button "Login"
        Then I wait to see the text "Welcome {username} !"
         """,
    )


# TODO:
# - add a step that takes a table and prints it
# - add a step that takes a multiline string and prints it
# - add a step that looks up a html element by css selector
# - add a step that waits with retries



# -*- coding: utf-8 -*-
#
# adapted from:
#   https://github.com/behave/behave/blob/master/behave/formatter/json.py
#
from __future__ import absolute_import

import json
import traceback
import uuid

import six
from behave.formatter.base import Formatter
from behave.model_core import Status
from tenacity import RetryError

from cucu.config import CONFIG


# -----------------------------------------------------------------------------
# CLASS: JSONFormatter
# -----------------------------------------------------------------------------
class CucuJSONFormatter(Formatter):
    name = "cucu-json"
    description = "JSON dump of test run"
    dumps_kwargs = {"indent": 2, "sort_keys": True}

    split_text_into_lines = True  # EXPERIMENT for better readability.

    json_number_types = six.integer_types + (float,)
    json_scalar_types = json_number_types + (six.text_type, bool, type(None))

    def __init__(self, stream_opener, config):
        super(CucuJSONFormatter, self).__init__(stream_opener, config)
        # -- ENSURE: Output stream is open.
        self.stream = self.open()
        self.feature_count = 0
        self.current_feature = None
        self.current_feature_data = None
        self.current_scenario = None
        self.last_step = None
        self.steps = []
        self.write_json_header()

    def reset(self):
        self.current_feature = None
        self.current_feature_data = None
        self.current_scenario = None

    # -- FORMATTER API:
    def uri(self, uri):
        pass

    def feature(self, feature):
        self.reset()
        self.current_feature = feature
        self.current_feature_data = {
            "keyword": feature.keyword,
            "name": feature.name,
            "tags": list(feature.tags),
            "location": six.text_type(feature.location),
            "status": None,  # Not known before feature run.
        }
        element = self.current_feature_data
        if feature.description:
            element["description"] = feature.description

    def background(self, background):
        element = self.add_feature_element(
            {
                "type": "background",
                "keyword": background.keyword,
                "name": background.name,
                "location": six.text_type(background.location),
                "steps": [],
            }
        )
        if background.name:
            element["name"] = background.name

        # -- ADD BACKGROUND STEPS: Support *.feature file regeneration.
        for step_ in background.steps:
            self.step(step_)

    def scenario(self, scenario):
        self.finish_current_scenario()
        self.current_scenario = scenario

        element = self.add_feature_element(
            {
                "type": "scenario",
                "keyword": scenario.keyword,
                "name": scenario.name,
                "tags": scenario.tags,
                "location": six.text_type(scenario.location),
                "steps": [],
                "status": None,
            }
        )
        self.steps = []
        if scenario.description:
            element["description"] = scenario.description

    @classmethod
    def make_table(cls, table):
        table_data = {
            "headings": table.headings,
            "rows": [list(row) for row in table.rows],
        }
        return table_data

    def insert_step(self, step, index=-1):
        step.unique_id = uuid.uuid1()
        is_substep = getattr(step, "is_substep", False)

        step_details = {
            "keyword": step.keyword,
            "step_type": step.step_type,
            "name": step.name,
            "location": six.text_type(step.location),
            "substep": is_substep,
        }

        if step.text:
            text = step.text
            if self.split_text_into_lines and "\n" in text:
                text = text.splitlines()
            step_details["text"] = text
        if step.table:
            step_details["table"] = self.make_table(step.table)
        element = self.current_feature_element

        if index == -1:
            self.steps.append(step)
            element["steps"].append(step_details)
        else:
            self.steps.insert(index, step)
            element["steps"].insert(index, step_details)

    def step(self, step):
        self.insert_step(step, index=-1)

    def match(self, match):
        # nothing to do, but we need to implement the method
        pass

    def result(self, step):
        steps = self.current_feature_element["steps"]
        step_index = 0
        for other_step in self.steps:
            if other_step.unique_id == step.unique_id:
                break
            step_index += 1

        # keep the last step recorded result state
        self.last_step = steps[step_index]

        timestamp = None
        if step.status.name in ["passed", "failed"]:
            timestamp = step.start_at

            step_variables = CONFIG.expand(step.name)

            if step.text:
                step_variables.update(CONFIG.expand(step.text))

            if step.table:
                for row in step.table.original.rows:
                    for value in row:
                        step_variables.update(CONFIG.expand(value))

            if step_variables:
                expanded = " ".join(
                    [
                        f'{key}="{value}"'
                        for (key, value) in step_variables.items()
                    ]
                )
                padding = f"    {' ' * (len('Given') - len(step.keyword))}"
                step.stdout.insert(
                    0, f"{padding}# {CONFIG.hide_secrets(expanded)}\n"
                )

        stdout = None
        if "stdout" in step.__dict__ and step.stdout != []:
            stdout = [CONFIG.hide_secrets("".join(step.stdout).rstrip())]

        stderr = None
        if "stderr" in step.__dict__ and step.stderr != []:
            stderr = [CONFIG.hide_secrets("".join(step.stderr).rstrip())]

        steps[step_index]["result"] = {
            "stdout": stdout,
            "stderr": stderr,
            "status": step.status.name,
            "duration": step.duration,
            "timestamp": timestamp,
        }

        if step.error_message and step.status == Status.failed:
            # -- OPTIONAL: Provided for failed steps.
            error_message = CONFIG.hide_secrets(step.error_message)
            if self.split_text_into_lines:
                error_message = error_message.splitlines()

            result_element = steps[step_index]["result"]
            result_element["error_message"] = error_message

            if error := step.exception:
                if isinstance(error, RetryError):
                    error = error.last_attempt.exception()

                if len(error.args) > 0 and isinstance(error.args[0], str):
                    error_class_name = error.__class__.__name__
                    redacted_error_msg = CONFIG.hide_secrets(error.args[0])
                    error_lines = redacted_error_msg.splitlines()
                    error_lines[0] = f"{error_class_name}: {error_lines[0]}"
                else:
                    error_lines = [repr(error)]

                result_element["exception"] = error_lines

    def embedding(self, mime_type, data):
        # nothing to do, but we need to implement the method
        pass

    def eof(self):
        """
        End of feature
        """
        if not self.current_feature_data:
            return

        # -- NORMAL CASE: Write collected data of current feature.
        self.finish_current_scenario()
        self.update_status_data()

        self.write_json_feature(self.current_feature_data)
        self.reset()

    def close(self):
        self.write_json_footer()
        self.close_stream()

    # -- JSON-DATA COLLECTION:
    def add_feature_element(self, element):
        assert self.current_feature_data is not None
        if "elements" not in self.current_feature_data:
            self.current_feature_data["elements"] = []
        self.current_feature_data["elements"].append(element)
        return element

    @property
    def current_feature_element(self):
        assert self.current_feature_data is not None
        return self.current_feature_data["elements"][-1]

    def update_status_data(self):
        assert self.current_feature
        assert self.current_feature_data
        self.current_feature_data["status"] = self.current_feature.status.name

    def finish_current_scenario(self):
        if self.current_scenario:
            hook_failed = self.current_scenario.hook_failed
            if hook_failed:
                status_name = "error"
            else:
                status_name = self.current_scenario.status.name

            self.current_feature_element["status"] = status_name

            if status_name in ["failed", "error"]:
                # we need to record the error_message and exc_traceback in the
                # last executed step and mark it as failed so the reporting can
                # show the result correctly
                error_message = [self.current_scenario.error_message]
                error_message += traceback.format_tb(
                    self.current_scenario.exc_traceback
                )
                #  If a before scenario hook fails, last_step will be None.
                if (
                    self.last_step is not None
                    and "error_message" not in self.last_step["result"]
                ):
                    self.last_step["result"]["error_message"] = error_message

    # -- JSON-WRITER:
    def write_json_header(self):
        self.stream.write("[\n")

    def write_json_footer(self):
        self.stream.write("\n]\n")

    def write_json_feature(self, feature_data):
        if "elements" not in feature_data:
            return

        filtered_scenarios = [
            x
            for x in feature_data["elements"]
            if x["keyword"] == "Scenario"
            and (
                CONFIG["CUCU_SHOW_SKIPS"] == "true" or x["status"] != "skipped"
            )
        ]

        if len(filtered_scenarios) == 0:
            return

        feature_data["elements"] = filtered_scenarios

        if self.feature_count != 0:
            self.write_json_feature_separator()

        self.stream.write(json.dumps(feature_data, **self.dumps_kwargs))
        self.stream.flush()
        self.feature_count += 1

    def write_json_feature_separator(self):
        self.stream.write(",\n\n")



import re

from cucu import config


def search(regex, value):
    """
    search for a matching regex within the value provided.
    """
    match = re.search(regex, value)

    if match is None:
        raise AssertionError(f'"{regex}" did not match anything in "{value}"')


def search_and_save(regex, value, name, variable):
    """
    search for a matching regex within the value provided and then save the value
    of the group name provided to the variable name provided.
    """
    match = re.search(regex, value)

    if match is None:
        raise AssertionError(f'"{regex}" did not match anything in "{value}"')

    groups = match.groupdict()

    if name in groups:
        config.CONFIG[variable] = groups[name]

    else:
        raise AssertionError(
            f'group "{name}" not found when searching for regex "{regex}" in "{value}"'
        )


def match_and_save(regex, value, name, variable):
    """
    match the regex in the value provided and then save the value of the group
    name provided to the variable name provided.
    """
    match = re.match(regex, value)

    if match is None:
        raise AssertionError(f'"{regex}" did not match "{value}"')

    groups = match.groupdict()

    if name in groups:
        config.CONFIG[variable] = groups[name]

    else:
        raise AssertionError(
            f'group "{name}" not found when matching regex "{regex}" to "{value}"'
        )



from cucu import fuzzy, helpers
from cucu.utils import take_saw_element_screenshot

from . import base_steps


def find_tab(ctx, name, index=0):
    """
    find a tab containing the text provide.

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired tab on screen
      index(str):  the index of the tab if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()
    element = fuzzy.find(ctx.browser, name, ['*[role="tab"]'], index=index)

    take_saw_element_screenshot(ctx, "tab", name, index, element)

    return element


def click_tab(ctx, tab):
    """
    internal method to click a tab
    """
    ctx.check_browser_initialized()

    if base_steps.is_disabled(tab):
        raise AssertionError("unable to click the tab, as it is disabled")

    ctx.browser.click(tab)


def is_selected(tab):
    """
    internal method to check if a tab is currently selected
    """
    return tab.get_attribute("aria-selected") == "true"


def is_not_selected(tab):
    """
    internal method to check if a tab is currently not selected
    """
    return not is_selected(tab)


helpers.define_should_see_thing_with_name_steps("tab", find_tab)
helpers.define_action_on_thing_with_name_steps(
    "tab", "click", find_tab, click_tab
)
helpers.define_thing_with_name_in_state_steps(
    "tab", "selected", find_tab, is_selected
)
helpers.define_thing_with_name_in_state_steps(
    "tab", "not selected", find_tab, is_not_selected
)
helpers.define_thing_with_name_in_state_steps(
    "tab", "disabled", find_tab, base_steps.is_disabled
)
helpers.define_thing_with_name_in_state_steps(
    "tab", "not disabled", find_tab, base_steps.is_not_disabled
)



from cucu.browser.core import Browser


def search_in_all_frames(browser, search_function):
    """
    search all frames on the page for an element

    Warning: This leaves the browser in whatever frame was last searched so that
    users of this method are in that frame.

    parameters:
      browser           - the cucu.browser.Browser object
      search_function   - function to search for the element (within a frame)
    returns:
        the WebElement that matches (if found)
    """
    # check whatever frame we're currently in
    result = search_function()

    if not result:
        # we might have not been in the default frame so check again
        browser.switch_to_default_frame()

        result = search_function()
        if result:
            return result

        frames = browser.execute('return document.querySelectorAll("iframe");')
        for frame in frames:
            # need to be in the default frame in order to switch to a child
            # frame w/o getting a stale element exception
            browser.switch_to_default_frame()
            browser.switch_to_frame(frame)
            result = search_function()

            if result:
                return result

    return result


def run_in_all_frames(browser, search_function):
    """
    run the search function on all of the available frames and return the
    appending of all the arrays.

    Warning: This leaves the browser in whatever frame was last searched so that
    users of this method are in that frame.

    parameters:
      browser           - the cucu.browser.Browser object
      search_function   - function that returns an array of WebElements
    returns:
        the array of all of the WebElements found.
    """
    result = []

    browser.switch_to_default_frame()
    result += search_function()

    frames = browser.execute('return document.querySelectorAll("iframe");')
    for frame in frames:
        # need to be in the default frame in order to switch to a child
        # frame w/o getting a stale element exception
        browser.switch_to_default_frame()
        browser.switch_to_frame(frame)
        result += search_function()

    return result


def try_in_frames_until_success(browser: Browser, function_to_run) -> None:
    """
    Run the function on all of the possible frames one by one. It terminates
    if the function doesn't raise an exception on a frame.

    Warning: This leaves the browser in whatever frame the function is run successfully
    so that users of the this method are in that frame.

    Args:
        browser (Browser): the browser session
        function_to_run : a function that raises an exception if it fails

    Raises:
        AssertionError: when the function fails in all frames
    """
    browser.switch_to_default_frame()
    try:
        function_to_run()
    except Exception:
        frames = browser.execute('return document.querySelectorAll("iframe");')
        for frame in frames:
            # need to be in the default frame in order to switch to a child
            # frame w/o getting a stale element exception
            browser.switch_to_default_frame()
            browser.switch_to_frame(frame)
            try:
                function_to_run()
            except Exception:
                if frames.index(frame) < len(frames) - 1:
                    continue
                else:
                    raise AssertionError(
                        f"{function_to_run.__name__} failed in all frames"
                    )
            return



# Define custom before/after hooks here
# Warn: Don't import this file into other python files to avoid duplicating hook registers

import datetime

from cucu import (
    register_custom_scenario_subheader_in_report_handling,
    register_custom_tags_in_report_handling,
)
from cucu.config import CONFIG
from cucu.environment import *  # noqa

# point the `cucu lint` command to our own custom linting rules
CONFIG["CUCU_LINT_RULES_PATH"] = "features/lint_rules"


def make_scenario_id_clickable(tag):
    """Example of processing a tag to open up a link to some url"""
    scenario_id = tag.replace(r"@sid-", "")
    return f'<a href="https://localhost/index.php?/cases/view/{scenario_id}" target="_blank">{tag}</a>'


register_custom_tags_in_report_handling(r"@sid-.+", make_scenario_id_clickable)


def my_example_subheader(scenario):
    """Example of a custom subheader for the report

    This may be useful if you want to link results to your log monitor system
    """

    query_time = ""
    if (
        "steps" in scenario
        and len(scenario["steps"]) > 0
        and "result" in scenario["steps"][0]
    ):
        one_minute = datetime.timedelta(minutes=1)
        # get start time from first step iso timestamp
        start_datetime = (
            datetime.datetime.fromisoformat(
                scenario["steps"][0]["result"]["timestamp"]
            )
            - one_minute
        )
        epoc_start_time_ms = int(start_datetime.timestamp()) * 1000

        end_datetime = start_datetime + (
            one_minute * 30
        )  # default to start + 30 minutes
        for step in reversed(scenario["steps"]):
            if step.get("result", []) and step["result"].get(
                "timestamp", None
            ):
                end_duration = (
                    datetime.timedelta(
                        seconds=step["result"].get("duration", 0)
                    )
                    + one_minute
                )
                end_datetime = (
                    datetime.datetime.fromisoformat(
                        step["result"]["timestamp"]
                    )
                    + end_duration
                )
                break

        epoc_end_time_ms = int(end_datetime.timestamp()) * 1000
        query_time = f"&begin={epoc_start_time_ms}&end={epoc_end_time_ms}"

    account_id = "123456"
    result = ""
    if account_id and query_time:
        html = "<span>Log Monitoring System: "
        html += f'<a href="https://localhost/logger?account={account_id}{query_time}" target="_blank">Link 🔗</a> '
        result = html

    return result


register_custom_scenario_subheader_in_report_handling(my_example_subheader)


# TODO:
# - add a register_page_check_hook  # to ensure the page is loaded)
# - add a register_before_retry_hook  # when using the retry decorator on steps)
# - add a register_custom_tags_in_report_handling
# - add a register_custom_scenario_subheader_in_report_handling
# - add a register_custom_variable_handling



from selenium.webdriver.support.ui import WebDriverWait

from cucu import fuzzy, helpers, logger
from cucu.utils import take_saw_element_screenshot


def find_draggable_element(ctx, name, index=0):
    """
    find a draggable element on screen by fuzzy matching on the name provided
    and the target element:

        * <* draggable="true">

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired draggable element on screen
      index(str):  the index of the draggable element if there are duplicates
    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()
    element = fuzzy.find(
        ctx.browser, name, ['*[draggable="true"]'], index=index
    )

    take_saw_element_screenshot(ctx, "draggable", name, index, element)

    return element


def is_draggable(element):
    """
    Checks if an element is draggable.
    Args:
        element (WebElement): The element to check.
    Returns:
        bool: True if the element is draggable, False otherwise.
    """
    return element.get_attribute("draggable") == "true"


def is_not_draggable(element):
    """
    Checks if an element is not draggable.
    Args:
        element (WebElement): The element to check.
    Returns:
        bool: True if the element is not draggable, False otherwise.
    """
    return not is_draggable(element)


def find_target_element(ctx, name, index=0):
    ctx.check_browser_initialized()
    _element = fuzzy.find(ctx.browser, name, ["*"], index=index)

    return _element


JS_DRAG_AND_DROP = """
    const cucuDragAndDrop = async (dragElem, dropElem) => {

    function triggerEvent(elem, eventName, dataTransfer = null) {
        return new Promise((resolve) => {
            const eventObj = new DragEvent(eventName, {
                bubbles: true,
                cancelable: true,
                composed: true,
                dataTransfer: dataTransfer,
            });

            function listener() {
                resolve(eventObj);
                elem.removeEventListener(eventName, listener);
            }

            elem.addEventListener(eventName, listener);
            elem.dispatchEvent(eventObj);
        });
    }

    const dragstartObj = await triggerEvent(dragElem, 'dragstart', new DataTransfer());
    await triggerEvent(dropElem, 'dragenter', dragstartObj.dataTransfer);
    const dropeventObj = await triggerEvent(dropElem, 'drop', dragstartObj.dataTransfer);
    await triggerEvent(dragElem, 'dragend', dropeventObj.dataTransfer);

    window.dragAndDropCompleted = true;
    };

    cucuDragAndDrop(arguments[0], arguments[1]);
"""


def drag_element_to_element(ctx, drag_name, drop_name):
    driver = ctx.browser.driver

    driver.execute_script("window.dragAndDropCompleted = false;")

    start_drag_rect = drag_name.rect
    logger.debug(
        f"Start location of drag element {drag_name.text}: {start_drag_rect}"
    )
    logger.debug("Executing drag-and-drop via JavaScript.")

    driver.execute_script(JS_DRAG_AND_DROP, drag_name, drop_name)

    # Wait for the JavaScript flag to be set to True
    WebDriverWait(driver, 10).until(
        lambda driver: driver.execute_script(
            "return window.dragAndDropCompleted;"
        )
    )

    end_drag_rect = drag_name.rect
    logger.debug(
        f"End location of drag element {drag_name.text}: {end_drag_rect}"
    )

    if start_drag_rect == end_drag_rect:
        raise AssertionError(
            f"Drag element {drag_name.text} position did not change"
        )
    else:
        logger.debug("Drag element position changed successfully.")

    logger.debug("Drag-and-drop operation executed.")


helpers.define_thing_with_name_in_state_steps(
    "element", "draggable", find_target_element, is_draggable
)

helpers.define_thing_with_name_in_state_steps(
    "element", "not draggable", find_target_element, is_not_draggable
)

helpers.define_thing_with_name_in_state_steps(
    "element", "draggable", find_target_element, is_draggable, with_nth=True
)

helpers.define_thing_with_name_in_state_steps(
    "element",
    "not draggable",
    find_target_element,
    is_not_draggable,
    with_nth=True,
)

helpers.define_two_thing_interaction_steps(
    "drag",
    drag_element_to_element,
    "element",
    find_draggable_element,
    "to",
    "element",
    find_target_element,
)

helpers.define_two_thing_interaction_steps(
    "drag",
    drag_element_to_element,
    "element",
    find_draggable_element,
    "to",
    "element",
    find_target_element,
    with_nth=True,
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import os
import time
from pathlib import Path

from behave.formatter.base import Formatter
from behave.model import ScenarioOutline, Status

from cucu import logger
from cucu.config import CONFIG
from cucu.db import (
    close_db,
    create_database_file,
    finish_cucu_run_record,
    finish_feature_record,
    finish_scenario_record,
    finish_step_record,
    finish_worker_record,
    record_cucu_run,
    record_feature,
    record_scenario,
    start_step_record,
)
from cucu.utils import (
    generate_short_id,
    get_iso_timestamp_with_ms,
)


class RundbFormatter(Formatter):
    """
    Record to the database using the Formatter Behave API.
    This is different from hooks as we don't have access to context (ctx) here.
    Instead we use this class's data and add some data to the passed in objects: feature, scenario, step.
    Another difference is that we can't rely on hooks to execute, as the step/scenario/feature may be skipped.
    Also, use cucu's CONFIG for global data since we don't have context (ctx).

    ## Just for reference:
    Processing Logic (simplified, without ScenarioOutline and skip logic)::

        for feature in runner.features:
            formatter = make_formatters(...)
            formatter.uri(feature.filename)
            formatter.feature(feature)
            for scenario in feature.scenarios:
                formatter.scenario(scenario)
                for step in scenario.all_steps:
                    formatter.step(step)
                    step_match = step_registry.find_match(step)
                    formatter.match(step_match)
                    if step_match:
                        step_match.run()
                    else:
                        step.status = Status.undefined
                    formatter.result(step.status)
            formatter.eof() # -- FEATURE-END
        formatter.close()
    """

    # -- FORMATTER API:
    name = "rundb"
    description = "records the results of the run to the run.db database"

    def __init__(self, stream_opener, config):
        super(RundbFormatter, self).__init__(stream_opener, config)
        # We don't actually use the stream provided by Behave, so don't open it.
        # self.stream = self.open()
        self.config = config

        if (
            not CONFIG["RUN_DB_PATH"]
            or CONFIG["WORKER_RUN_ID"] != CONFIG["WORKER_PARENT_ID"]
        ):
            logger.debug(
                "Create a new worker db since this isn't the parent process"
            )
            # use seed unique enough for multiple cucu_runs to be combined but predictable within the same run
            worker_id_seed = f"{CONFIG['WORKER_PARENT_ID']}_{os.getpid()}"
            CONFIG["WORKER_RUN_ID"] = generate_short_id(worker_id_seed)

            results_path = Path(CONFIG["CUCU_RESULTS_DIR"])
            worker_run_id = CONFIG["WORKER_RUN_ID"]
            cucu_run_id = CONFIG["CUCU_RUN_ID"]
            CONFIG["RUN_DB_PATH"] = run_db_path = (
                results_path / f"run_{cucu_run_id}_{worker_run_id}.db"
            )

            if not run_db_path.exists():
                logger.debug(
                    f"Creating new run database file: {run_db_path} for {worker_id_seed}"
                )
                create_database_file(run_db_path)
                record_cucu_run()

    def uri(self, uri):
        # nothing to do, but we need to implement the method for behave
        pass

    def feature(self, feature):
        """Called before a feature is executed."""
        self.this_feature = feature
        self.this_scenario = None
        self.this_background = None
        self.next_start_at = None

        feature_run_id_seed = (
            f"{CONFIG['WORKER_RUN_ID']}_{time.perf_counter()}"
        )
        feature.feature_run_id = generate_short_id(feature_run_id_seed)
        feature.custom_data = {}

        record_feature(feature)

    def background(self, background):
        # nothing to do, but we need to implement the method for behave
        pass

    def _finish_scenario(self):
        if self.this_scenario is None:
            return

        # ensure non-executed steps have correct seq
        for index, step in enumerate(self.this_steps):
            if getattr(step, "seq", -1) == -1:
                step.seq = index + 1  # 1-based sequence
                finish_step_record(step, 0)

        finish_scenario_record(self.this_scenario)

    def scenario(self, scenario):
        """Called before a scenario is executed (or ScenarioOutline scenarios)."""
        self._finish_scenario()

        self.this_scenario = scenario
        self.this_steps = []
        self.next_start_at = get_iso_timestamp_with_ms()
        scenario.custom_data = {}
        scenario.scenario_run_id = generate_short_id(
            f"{scenario.feature.feature_run_id}_{time.perf_counter()}"
        )

        # feature.scenarios is a mix of Scenario and ScenarioOutline objects with their own scenarios list
        for index, feature_scenario in enumerate(scenario.feature.scenarios):
            if feature_scenario == scenario:
                scenario.seq = index + 1
                break

            # Scenarios belonging to a Scenario Outline are included under their Scenario Outline sequence.
            # Add a suffix to preserve ordering within the Scenario Outline.
            if isinstance(feature_scenario, ScenarioOutline):
                for sub_index, sub_scenario in enumerate(
                    feature_scenario._scenarios
                ):
                    if sub_scenario == scenario:
                        scenario.seq = index + 1 + (sub_index + 1) / 10
                        break

        record_scenario(scenario)

    def step(self, step):
        """Called before a step is executed (and matched)."""
        self.insert_step(step, index=-1)

    def insert_step(self, step, index):
        """cucu specific step insertion method used to add steps here and dynamically"""
        next_index = index if index != -1 else len(self.this_steps)
        step_run_id_seed = f"{self.this_scenario.scenario_run_id}_{next_index}_{time.perf_counter()}"
        step.step_run_id = generate_short_id(
            step_run_id_seed, length=10
        )  # up to 10 characters to give two orders of magnitude less chance of collision
        self.this_steps.insert(next_index, step)
        start_step_record(step, self.this_scenario.scenario_run_id)

    def match(self, match):
        # nothing to do, but we need to implement the method for behave
        pass

    def result(self, step):
        """Called after processing a step result is known, applies to executed/skipped too."""
        step.start_at = self.next_start_at
        self.next_start_at = step.end_at = get_iso_timestamp_with_ms()
        previous_step_duration = getattr(
            self.this_scenario, "previous_step_duration", 0
        )
        if step.status in (Status.untested, Status.undefined):
            step.seq = self.this_steps.index(step) + 1

        # consider undefined as test author's failure, not a framework issue.
        if step.status == Status.undefined:
            step.status = Status.failed
            step.error_message = "Step is undefined"

        finish_step_record(step, previous_step_duration)

    def eof(self):
        """Called after processing a feature (or a feature file)."""
        # need to finish the last scenario
        self._finish_scenario()
        finish_feature_record(self.this_feature)

    def close(self):
        """Called before the formatter is no longer used
        (stream/io compatibility).
        """
        finish_worker_record(None)
        finish_cucu_run_record()
        close_db()



#
# cucu's matching sub system which allows for a pluggable matching architecture
#

# WIP:  thinking of how to generalize the idea of finding elements on the
#       currently opened browser while also being able to handle opening
#       browsers through selenium, cypress or whatever.


class ElementMatcher:
    def __init__(self):
        pass

    def find_button(self, execute_script, name, index: 0):
        """
        return the button labeled by the name provided and use the index to
        pick a specific one if there are duplicates.

        arguments:
            execute_script - function to execute javascript in the currently
                             running browsers javascript console.
            name           - name of the button we want to find on screen.
            index          - 0-base index of which element to return if there
                             are duplicates. default: 0

        return:
            return the Element found

        """
        raise RuntimeError("implement me")



from behave.formatter.base import Formatter


class CucuTagCollectorFormatter(Formatter):
    name = "tags.filter"
    description = "list tags that match a certain regular expression"

    def __init__(self, stream_opener, config):
        super(CucuTagCollectorFormatter, self).__init__(stream_opener, config)
        self.tags = set()

    def uri(self, uri):
        pass

    def feature(self, feature):
        if feature.should_run(config=self.config):
            [self.tags.add(tag) for tag in feature.tags]

    def scenario(self, scenario):
        if scenario.should_run(config=self.config):
            [self.tags.add(tag) for tag in scenario.tags]

    def close(self):
        self.open()
        self.stream.write("\n".join(sorted(self.tags)) + "\n")
        self.close_stream()



#
# XXX: a lot of convoluted logic to handle capturing per step stdou/stderr
#      writing while not interfering with the way behave does its own log
#      capturing
#
import os
import re
import sys
import warnings
from functools import wraps

import parse

# set env var BEHAVE_STRIP_STEPS_WITH_TRAILING_COLON=yes before importing behave
os.environ["BEHAVE_STRIP_STEPS_WITH_TRAILING_COLON"] = "yes"

import behave
from behave.__main__ import main as original_behave_main
from behave.model import Table

from cucu.config import CONFIG


def behave_main(args):
    return original_behave_main(args)


def init_outputs(stdout, stderr):
    # capturing stdout and stderr output to record in reporting
    sys.stdout = CucuOutputStream(sys.stdout)
    sys.stderr = CucuOutputStream(sys.stderr)


NTH_REGEX = r"(\d+)(nd|th|rd|st)"


@parse.with_pattern(NTH_REGEX)
def parse_nth(nth):
    matcher = re.match(NTH_REGEX, nth)

    if matcher is None:
        raise Exception(f"nth expression {nth} is invalid")

    number, _ = matcher.groups()
    return int(number) - 1


def init_step_hooks(stdout, stderr):
    init_outputs(stdout, stderr)
    #
    # wrap the given, when, then, step decorators from behave so we can intercept
    # the step arguments and do things such as replacing variable references that
    # are wrapped with curly braces {...}.
    #
    for decorator_name in ["given", "when", "then", "step"]:
        decorator = behave.__dict__[decorator_name]

        def inner_step_func(func, *args, variable_passthru=False, **kwargs):
            #
            # replace the variable references in the args and kwargs passed to the
            # step
            #
            # TODO: for tables and multiline strings there's a bit more
            #       work to be done here.
            #
            if not variable_passthru:
                args = [CONFIG.resolve(value) for value in args]
                kwargs = {
                    key: CONFIG.resolve(val) for (key, val) in kwargs.items()
                }

                # resolve variables in text and table data
                ctx = args[0]

                # we know what we're doing modifying this ctx value and so lets
                # avoid the unnecessary warning int he logs
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")

                    # resolve variables in the multiline string arguments
                    ctx.text = CONFIG.resolve(ctx.text)

                    # resolve variables in the table values
                    if ctx.table is not None:
                        ctx.table.original = Table(
                            ctx.table.headings,
                            rows=ctx.table.rows,
                            line=ctx.table.line,
                        )
                        new_rows = []
                        for row in ctx.table.rows:
                            new_row = []

                            for value in row:
                                new_row.append(CONFIG.resolve(value))

                            new_rows.append(new_row)

                        ctx.table.rows = new_rows

            func(*args, **kwargs)

        def new_decorator(
            step_text, fix_inner_step=lambda x: x, variable_passthru=False
        ):
            """
            the new @step decorator

            parameters:
              step_text(string): the actual step string definition.
              fix_inner_step(function): a function used to fix the inner_step
                                        function code location and other
                                        properties.
              variable_passthru(bool): when set to true the variables are not
                                       replaced in the arguments of the step
                                       text and so variable references such as
                                       {FOO} are passed as such so they can
                                       be handled further down in `run_steps`
                                       for example.
            """

            # ensure register nth type after Runner reset when used with multiple workers
            parse_matcher = behave.matchers._the_step_matcher_factory.step_matcher_class_mapping[
                "parse"
            ]
            if not parse_matcher.has_registered_type(name="nth"):
                parse_matcher.register_type(nth=parse_nth)

            #
            # IMPORTANT: if you add any additional line from the `wrapper`
            #            function name to the `inner_step` name then you HAVE to
            #            update the `src/cucu/helpers.py` fix_inner_step
            #            so it subtracts the lines between wrapper and
            #            inner_step functions (lines of code). More details
            #            at the `src/cucu/helpers.py` location.
            #
            def wrapper(func):
                @decorator(step_text)
                @wraps(func)
                def inner_step(*args, **kwargs):
                    inner_step_func(
                        func,
                        *args,
                        variable_passthru=variable_passthru,
                        **kwargs,
                    )

                fix_inner_step(inner_step)

            return wrapper

        behave.__dict__[decorator_name.title()] = behave.__dict__[
            decorator_name
        ] = new_decorator


_stdout = sys.stdout


class CucuOutputStream:
    """
    encapsulates a lot of the logic to handle capturing step by step console
    logging but also redirecting logging at runtime to another stream


    FYI: in order to print to the screen you must use directly the sys.stdout object
         and write to it here as doing something like `print(...)` will result
         in an infinite recursive loop and break
    """

    def __init__(self, stream, other_stream=None):
        self.captured_data = []
        self.stream = stream
        self.encoding = stream.encoding
        self.other_stream = other_stream

    def set_other_stream(self, other):
        self.other_stream = other

    def __getattr__(self, item):
        return self.stream.__getattribute__(item)

    def isatty(self, *args, **kwargs):
        return self.stream.isatty(*args, **kwargs)

    def fileno(self):
        return self.stream.fileno()

    def flush(self):
        self.stream.flush()

        if self.other_stream:
            self.other_stream.flush()

    def write(self, byte):
        byte = CONFIG.hide_secrets(byte)

        self.captured_data.append(byte)

        if isinstance(byte, bytes):
            self.stream.write(byte.decode("utf8"))

            if self.other_stream:
                self.other_stream.write(byte.decode("utf8"))
        else:
            self.stream.write(byte)

            if self.other_stream:
                self.other_stream.write(byte)

    def writelines(self, lines):
        lines = [CONFIG.hide_secrets(line) for line in lines]

        for line in lines:
            self.captured_data.append(line)

        self.stream.writelines(lines)
        if self.other_stream:
            self.other_stream.writelines(lines)

    def captured(self):
        """
        returns the data captured thus far to the stream and resets the internal
        buffer to empty.
        """
        captured_data = self.captured_data
        self.captured_data = []
        return captured_data



from cucu import fuzzy, helpers, retry, step
from cucu.config import CONFIG
from cucu.utils import (
    find_n_click_input_parent_label,
    is_element_size_zero,
    take_saw_element_screenshot,
)

from . import base_steps


def find_radio_button(ctx, name, index=0):
    """
    find a radio button on screen by fuzzy matching on the name provided and
    the target element:

        * <input type="radio">
        * <* role="radio">

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired radio button on screen
      index(str):  the index of the radio button if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()
    element = fuzzy.find(
        ctx.browser,
        name,
        [
            'input[type="radio"]',
            '*[role="radio"]',
        ],
        index=index,
        direction=fuzzy.Direction.RIGHT_TO_LEFT,
    )

    take_saw_element_screenshot(ctx, "radio button", name, index, element)

    return element


def find_n_assert_radio_button(ctx, name, index=0, is_visible=True):
    """
    find and assert a radio button is visible

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired radio button on screen
      index(str):  the index of the radio button if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    radio = find_radio_button(ctx, name, index=index)

    if is_visible:
        if radio is None:
            raise AssertionError(f'unable to find radio button "{name}"')
    else:
        if radio is not None:
            raise AssertionError(f'able to find radio button "{name}"')

    return radio


def find_n_select_radio_button(ctx, name, index=0, ignore_if_selected=False):
    """
    find and select a radio button

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired radio button on screen
      index(str):  the index of the radio button if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()
    radio = find_n_assert_radio_button(ctx, name, index=index)

    if base_steps.is_disabled(radio):
        raise AssertionError(
            "unable to select the radio button, as it is disabled"
        )

    selected = radio.get_attribute("checked") == "true"

    if selected:
        if ignore_if_selected:
            return

        raise AssertionError(f'radio button "{name}" already selected')

    if is_element_size_zero(radio):
        find_n_click_input_parent_label(ctx, radio)
        return

    ctx.browser.click(radio)


def is_selected(radio):
    """
    internal method to check if radiobox is selected
    """
    return bool(radio.get_attribute("checked"))


def is_not_selected(radio):
    """
    internal method to check if a radiobox is not selected
    """
    return not is_selected(radio)


def select_radio_button(ctx, radiobox):
    """
    internal method to select a radio button that isn't already selected
    """
    ctx.check_browser_initialized()

    if base_steps.is_disabled(radiobox):
        raise AssertionError(
            "unable to select the radio button, as it is disabled"
        )

    selected = bool(radiobox.get_attribute("checked"))

    if selected:
        raise AssertionError("radiobox already selected")

    if is_element_size_zero(radiobox):
        find_n_click_input_parent_label(ctx, radiobox)
        return

    ctx.browser.click(radiobox)


helpers.define_should_see_thing_with_name_steps(
    "radio button", find_radio_button
)
helpers.define_action_on_thing_with_name_steps(
    "radio button", "select", find_radio_button, select_radio_button
)
helpers.define_thing_with_name_in_state_steps(
    "radio button", "selected", find_radio_button, is_selected, with_nth=True
)
helpers.define_thing_with_name_in_state_steps(
    "radio button", "not selected", find_radio_button, is_not_selected
)
helpers.define_thing_with_name_in_state_steps(
    "radio button", "disabled", find_radio_button, base_steps.is_disabled
)
helpers.define_thing_with_name_in_state_steps(
    "radio button",
    "not disabled",
    find_radio_button,
    base_steps.is_not_disabled,
)
helpers.define_run_steps_if_I_can_see_element_with_name_steps(
    "radio", find_radio_button
)


@step('I immediately select the radio button "{name}" if it is not selected')
def immediately_select_the_radio_button_if_not_selected(ctx, name):
    find_n_select_radio_button(ctx, name, ignore_if_selected=True)


@step('I select the radio button "{name}" if it is not selected')
def select_the_radio_button_if_not_selected(ctx, name):
    retry(
        find_n_select_radio_button,
        retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
        wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
    )(ctx, name, ignore_if_selected=True)


@step('I wait to select the radio button "{name}" if it is not selected')
def wait_to_select_the_radio_button_if_not_selected(ctx, name):
    retry(find_n_select_radio_button)(ctx, name, ignore_if_selected=True)


@step(
    'I immediately select the "{nth:nth}" radio button "{name}" if it is not selected'
)
def immediately_select_nth_radio_button_if_not_selected(ctx, nth, name):
    find_n_select_radio_button(ctx, name, nth, ignore_if_selected=True)


@step('I select the "{nth:nth}" radio button "{name}" if it is not selected')
def select_nth_radio_button(ctx, nth, name):
    retry(
        find_n_select_radio_button,
        retry_after_s=float(CONFIG["CUCU_SHORT_UI_RETRY_AFTER_S"]),
        wait_up_to_s=float(CONFIG["CUCU_SHORT_UI_WAIT_TIMEOUT_S"]),
    )(ctx, name, nth, ignore_if_selected=True)



from selenium.webdriver.remote import webelement

from cucu.browser.frames import search_in_all_frames
from cucu.config import CONFIG

# monkey patch some methods at the WebElement level
__ORIGINAL_FIND_ELEMENTS = webelement.WebElement.find_elements


def init():
    """
    initialize various selenium tweaks to change the behavior of the underlying
    selenium classes.
    """

    # intercept the find_elements calls and use the nifty frames method to allow
    # the expression to be executed against every visible frame to find the
    # desired element.
    def find_elements(self, by="id", value=None):
        ctx = CONFIG["__CUCU_CTX"]

        def find_elements_in_frame():
            return __ORIGINAL_FIND_ELEMENTS(self, by, value)

        return search_in_all_frames(ctx.browser, find_elements_in_frame)

    webelement.WebElement.find_elements = find_elements



import os
import re

from cucu import retry, step
from cucu.config import CONFIG


@step('I create a file at "{filepath}" with the following')
def create_file_with_the_following(ctx, filepath):
    dirname = os.path.dirname(filepath)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(filepath, "wb") as output:
        output.write(bytes(ctx.text, "utf8"))


@step('I create the directory at "{filepath}"')
def create_directory_at(ctx, filepath):
    os.makedirs(filepath)


@step('I delete the file at "{filepath}"')
def delete_file_at(ctx, filepath):
    os.remove(filepath)


@step('I delete the file at "{filepath}" if it exists')
def delete_file_at_if_it_exists(ctx, filepath):
    if os.path.exists(filepath):
        os.remove(filepath)


@step(
    'I read the contents of the file at "{filepath}" and save to the variable "{variable}"'
)
def read_file_contents(ctx, filepath, variable):
    with open(filepath, "r") as _input:
        CONFIG[variable] = CONFIG.escape(_input.read())


@step('I append to the file at "{filepath}" the following')
def append_to_file_the_following(ctx, filepath):
    with open(filepath, "ab") as output:
        output.write(bytes(ctx.text, "utf8"))


def assert_file(ctx, filepath):
    if not (os.path.exists(filepath) and os.path.isfile(filepath)):
        raise AssertionError(f"unable to see file at {filepath}")


@step('I should see a file at "{filepath}"')
def should_see_file(ctx, filepath):
    assert_file(ctx, filepath)


@step('I should not see a file at "{filepath}"')
def should_not_see_file(ctx, filepath):
    if os.path.exists(filepath) and os.path.isfile(filepath):
        raise AssertionError(f"able to see file at {filepath}")


@step('I wait to see a file at "{filepath}"')
def wait_to_see_file(ctx, filepath):
    retry(assert_file)(ctx, filepath)


@step('I wait up to "{seconds}" seconds to see a file at "{filepath}"')
def wait_up_to_see_file(ctx, seconds, filepath):
    seconds = float(seconds)
    retry(assert_file, wait_up_to_s=seconds)(ctx, filepath)


@step('I should see the directory at "{filepath}"')
def should_see_directory(ctx, filepath):
    if not (os.path.exists(filepath) and os.path.isdir(filepath)):
        raise AssertionError(f"unable to see directory at {filepath}")


@step('I should not see the directory at "{filepath}"')
def should_not_see_directory(ctx, filepath):
    if os.path.exists(filepath) and os.path.isdir(filepath):
        raise AssertionError(f"able to see directory at {filepath}")


@step('I should see the file at "{filepath}" is equal to the following')
def should_see_file_is_equal_to_the_following(ctx, filepath):
    with open(filepath, "rb") as input:
        file_contents = input.read().decode("utf8")

        if file_contents != ctx.text:
            raise AssertionError(
                f"\n{file_contents}\n is not equal to:\n{ctx.text}\n"
            )


@step('I should see the file at "{filepath}" contains the following')
def should_see_file_contains_the_following(ctx, filepath):
    with open(filepath, "rb") as input:
        file_contents = input.read().decode("utf8")

        if ctx.text not in file_contents:
            raise AssertionError(
                f"\n{file_contents}\ndoes not contain\n{ctx.text}\n"
            )


@step('I should see the file at "{filepath}" contains the following pattern')
def should_see_file_contains_pattern(ctx, filepath):
    """Checks if the file contains text matching the given regex pattern."""
    with open(filepath, "rb") as input_file:
        file_contents = input_file.read().decode("utf-8")
        pattern = ctx.text.strip()
        anchored_pattern = rf"^\s*{pattern}\s*$"

        if not re.search(
            anchored_pattern, file_contents, re.MULTILINE | re.IGNORECASE
        ):
            raise AssertionError(
                f"\n{file_contents}\ndoes not match regex pattern:\n{pattern}\n"
            )


@step('I should see the file at "{filepath}" matches the following')
def should_see_file_matches_the_following(ctx, filepath):
    with open(filepath, "rb") as input:
        file_contents = input.read().decode("utf8")

        if not re.match(ctx.text, file_contents):
            raise AssertionError(
                f"\n{file_contents}\ndoes not match\n{ctx.text}\n"
            )


@step('I should see the file at "{filepath}" is not equal to the following')
def should_see_file_is_not_equal_to_the_following(ctx, filepath):
    with open(filepath, "rb") as input:
        file_contents = input.read().decode("utf8")

        if file_contents != ctx.text:
            raise AssertionError(
                f"\n{file_contents}\nis equal to\n{ctx.text}\n"
            )


@step('I should see the file at "{filepath}" does not contain the following')
def should_see_file_does_not_contain_the_following(ctx, filepath):
    with open(filepath, "rb") as input:
        file_contents = input.read().decode("utf8")

        if ctx.text in file_contents:
            raise AssertionError(f"\n{file_contents}\ncontains\n{ctx.text}\n")


@step('I should see the file at "{filepath}" does not match the following')
def should_see_file_does_not_match_the_following(ctx, filepath):
    with open(filepath, "rb") as input:
        file_contents = input.read().decode("utf8")

        if re.match(ctx.text, file_contents):
            raise AssertionError(f"\n{file_contents}\nmatches\n{ctx.text}\n")



import contextlib
import io

from cucu import behave_tweaks, init_global_hook_variables


def collect_cucu_tags(filepath=None, tags=tuple()):
    """
    Collects tests using behave and extracts the tags applied to them.

    Returns:
        A list of tags
    """
    args = [
        "--dry-run",
        "--no-summary",
        "--format",
        "cucu.formatter.tags:CucuTagCollectorFormatter",
    ]

    if len(tags) > 0:
        [args.extend(["--tags", tag]) for tag in tags]

    if filepath is not None:
        args.append(filepath)

    stdout = io.StringIO()
    stderr = io.StringIO()

    init_global_hook_variables()
    with contextlib.redirect_stderr(stderr):
        with contextlib.redirect_stdout(stdout):
            behave_tweaks.behave_main(args)

    stdout = [
        tag for tag in stdout.getvalue().split("\n") if len(tag.strip()) > 0
    ]

    if len(stdout) > 0 and stdout[0].startswith("USING RUNNER:"):
            # remove the first line of the output so that linter won't error out trying to parse it as a step
            stdout = stdout[1:]

    if len(stdout) > 0 and  stdout[0].startswith("ParserError"):
        raise RuntimeError(
            "Unable to parse feature files. Please double-check the syntax of your feature files."
            )
    return stdout



from cucu import fuzzy, helpers
from cucu.utils import take_saw_element_screenshot

from . import base_steps


def find_link(ctx, name, index=0):
    """
    find a link on screen by fuzzy matching on the name provided and the target
    element:

        * <a>
        * <* role="link">

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired link on screen
      index(str):  the index of the link if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()
    element = fuzzy.find(
        ctx.browser, name, ["a", '*[role="link"]'], index=index
    )

    take_saw_element_screenshot(ctx, "link", name, index, element)

    return element


def click_link(ctx, link):
    """
    internal method to click a link
    """
    ctx.check_browser_initialized()

    if base_steps.is_disabled(link):
        raise AssertionError("unable to click the link, as it is disabled")

    ctx.browser.click(link)


helpers.define_should_see_thing_with_name_steps(
    "link", find_link, with_nth=True
)
helpers.define_action_on_thing_with_name_steps(
    "link", "click", find_link, click_link, with_nth=True
)
helpers.define_thing_with_name_in_state_steps(
    "link", "disabled", find_link, base_steps.is_disabled, with_nth=True
)
helpers.define_thing_with_name_in_state_steps(
    "link",
    "not disabled",
    find_link,
    base_steps.is_not_disabled,
    with_nth=True,
)
helpers.define_run_steps_if_I_can_see_element_with_name_steps(
    "link", find_link
)



# coding: utf-8

import os
import logging
from . import utils


def install(cwd=False):
    """
    Appends the directory of the edgedriver binary file to PATH.

    :param cwd: Flag indicating whether to download to current working directory
    :return: The file path of edgedriver
    """
    edgedriver_filepath = utils.download_edgedriver(cwd)
    if not edgedriver_filepath:
        logging.debug("Can not download edgedriver.")
        return
    edgedriver_dir = os.path.dirname(edgedriver_filepath)
    if "PATH" not in os.environ:
        os.environ["PATH"] = edgedriver_dir
    elif edgedriver_dir not in os.environ["PATH"]:
        os.environ["PATH"] = (
            edgedriver_dir
            + utils.get_variable_separator()
            + os.environ["PATH"]
        )
    return edgedriver_filepath


def get_edge_version():
    """
    Get installed version of edge on client

    :return: The version of edge
    """
    return utils.get_edge_version()



import sys

import humanize

# XXX: this would have to be generalized to other browser abstractions
from selenium.webdriver.common.keys import Keys

from cucu import fuzzy, helpers, retry, step
from cucu.utils import take_saw_element_screenshot

from . import base_steps


def find_input(ctx, name, index=0):
    """
    find an input on screen by fuzzy matching on the name and index provided.

        * <input>
        * <textarea>

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired input on screen
      index(str):  the index of the input if there are a few with the same name.

    returns:
        the WebElement that matches the provided arguments.

    raises:
        an AssertionError if the input isn't found
    """
    ctx.check_browser_initialized()

    element = fuzzy.find(
        ctx.browser,
        name,
        [
            # we can only write into things that do not have the type:
            # button, checkbox, radio, color, hidden, range, reset
            "input[type!=button][type!=checkbox][type!=radio][button!=color][button!=hidden][button!=range][button!=reset]",
            "textarea",
        ],
        index=index,
    )

    prefix = "" if index == 0 else f"{humanize.ordinal(index + 1)} "

    take_saw_element_screenshot(ctx, "input", name, index, element)

    if element is None:
        raise AssertionError(f'unable to find the {prefix}input "{name}"')

    return element


def click_input(ctx, input_):
    """
    internal method used to simply click a input element
    """
    ctx.check_browser_initialized()

    if base_steps.is_disabled(input_):
        raise AssertionError("unable to click the input, as it is disabled")

    ctx.browser.click(input_)


def clear_input(input_):
    """
    internal method used to clear inputs and make sure they clear correctly
    """
    # Keys.CONTROL works on the Selenium grid (when running in CI)
    #  - it stopped working on laptop after the upgrade in antd version (Feb 2023)
    # Keys.COMMAND works on laptop, but not on Selenium grid,
    # and actually causes an active session on the grid to hang.
    if "darwin" in sys.platform:
        input_.send_keys(Keys.COMMAND, "a")
    else:
        input_.send_keys(Keys.CONTROL, "a")
    input_.send_keys(Keys.BACKSPACE)


def find_n_write(ctx, name, value, index=0, clear_existing=True):
    """
    find the input with the name provided and write the value provided into it.

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired input on screen
      index(str):  the index of the input if there are a few with the same name.
      clear_existing(bool): if clearing the existing content before writing

    raises:
        an error if the desired input is not found
    """
    ctx.check_browser_initialized()

    input_ = find_input(ctx, name, index=index)

    if base_steps.is_disabled(input_):
        raise AssertionError(
            "unable to write into the input, as it is disabled"
        )

    if clear_existing:
        clear_input(input_)

    if len(value) > 512:
        #
        # to avoid various bugs with writing large chunks of text into
        # inputs/textareas using send_keys we can just put the text into the
        # @value attribute and simulate a keystroke so other UI related events
        # fire, list of some bugs:
        #
        #  * https://github.com/seleniumhq/selenium-google-code-issue-archive/issues/4469
        #  * various stackoverflow articles on this matter
        #
        ctx.browser.execute(
            "arguments[0].value = arguments[1];", input_, value
        )
        input_.send_keys(" ")
        input_.send_keys(Keys.BACKSPACE)

    else:
        input_.send_keys(value)


def find_n_clear(ctx, name, index=0):
    """
    find the input with the name provided and clear whatever value it has.

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired input on screen
      index(str):  the index of the input if there are a few with the same name.

    raises:
        an error if the desired input is not found
    """
    ctx.check_browser_initialized()

    input_ = find_input(ctx, name, index=index)

    if base_steps.is_disabled(input_):
        raise AssertionError("unable to clear the input, as it is disabled")

    clear_input(input_)


def assert_input_value(ctx, name, value, index=0):
    """
    assert the input with the name provided has the value specified, unless the
    value is None which then means to verify there is not value set.

    Params:
        ctx(object):   behave context used to share information between steps
        name(string):  name of the input to find and assert is visible
        value(string): value to assert the input has, when set to None we verify
                       the input currently has no value.
        index(int):    index of the input to assert when there are multiple
                       inputs have the same name.
    """
    input_ = find_input(ctx, name, index=index)
    actual = input_.get_attribute("value")
    prefix = "" if index == 0 else f"{humanize.ordinal(index + 1)} "

    if value is None:
        if actual != "":
            raise AssertionError(
                f'the {prefix}input "{name}" has value "{actual}"'
            )

    elif actual != value:
        raise AssertionError(
            f'the {prefix}input "{name}" has value "{actual}"'
        )


helpers.define_should_see_thing_with_name_steps("input", find_input)
helpers.define_thing_with_name_in_state_steps(
    "input", "disabled", find_input, base_steps.is_disabled
)
helpers.define_thing_with_name_in_state_steps(
    "input", "not disabled", find_input, base_steps.is_not_disabled
)
helpers.define_run_steps_if_I_can_see_element_with_name_steps(
    "input", find_input
)
helpers.define_action_on_thing_with_name_steps(
    "input", "click", find_input, click_input, with_nth=True
)


@step('I write "{value}" into the input "{name}"')
def writes_into_input(ctx, value, name):
    find_n_write(ctx, name, value)


@step('I wait to write "{value}" into the input "{name}"')
def wait_to_write_into_input(ctx, value, name):
    retry(find_n_write)(ctx, name, value)


@step(
    'I wait up to "{seconds}" seconds to write "{value}" into the input "{name}"'
)
def wait_up_to_write_into_input(ctx, seconds, value, name):
    retry(find_n_write, wait_up_to_s=float(seconds))(ctx, name, value)


@step('I write the following into the input "{name}"')
def writes_multi_lines_into_input(ctx, name):
    find_n_write(ctx, name, ctx.text)


@step('I wait to write the following into the input "{name}"')
def wait_to_write_multi_lines_into_input(ctx, name):
    retry(find_n_write)(ctx, name, ctx.text)


@step(
    'I wait up to "{seconds}" to write the following into the input "{name}"'
)
def wait_up_to_write_multi_lines_into_input(ctx, seconds, name):
    retry(find_n_write, wait_up_to_s=float(seconds))(ctx, name, ctx.text)


@step('I send the "{key}" key to the input "{name}"')
def send_keys_to_input(ctx, key, name):
    find_n_write(ctx, name, Keys.__dict__[key.upper()], clear_existing=False)


@step('I clear the input "{name}"')
def clear_the_input(ctx, name):
    find_n_clear(ctx, name)


@step('I wait to clear the input "{name}"')
def wait_to_clear_input(ctx, name):
    retry(find_n_clear)(ctx, name)


@step('I wait up to "{seconds}" seconds to clear the input "{name}"')
def wait_up_to_clear_input(ctx, seconds, name):
    retry(find_n_clear, wait_up_to_s=float(seconds))(ctx, name)


@step('I should see the input "{name}" is empty')
def should_see_the_input_is_empty(ctx, name):
    assert_input_value(ctx, name, None)


@step('I should see "{value}" in the input "{name}"')
def should_see_the_input_with_value(ctx, value, name):
    assert_input_value(ctx, name, value)


@step('I wait to see the value "{value}" in the input "{name}"')
def wait_to_see_the_input_with_value(ctx, value, name):
    retry(assert_input_value)(ctx, name, value)


@step('I should see the following in the input "{name}"')
def should_see_the_following_input_with_value(ctx, name):
    assert_input_value(ctx, name, ctx.text)


@step('I wait to see the following in the input "{name}"')
def wait_to_see_the_following_input_with_value(ctx, name):
    retry(assert_input_value)(ctx, name, ctx.text)


@step('I should see no value in the input "{name}"')
def should_to_see_the_input_with_no_value(ctx, name):
    assert_input_value(ctx, name, None)


@step('I write "{value}" into the "{nth:nth}" input "{name}"')
def write_into_the_nth_input(ctx, value, nth, name):
    find_n_write(ctx, name, value, index=nth)


@step('I wait to write "{value}" into the "{nth:nth}" input "{name}"')
def wait_to_write_into_the_nth_input(ctx, value, nth, name):
    retry(find_n_write)(ctx, name, value, index=nth)


@step('I wait to see the "{nth:nth}" input "{name}"')
def wait_to_see_the_nth_input(ctx, nth, name):
    retry(find_input)(ctx, name, index=nth)


@step('I should see "{value}" in the "{nth:nth}" input "{name}"')
def should_see_the_nth_input_with_value(ctx, value, nth, name):
    assert_input_value(ctx, name, value, index=nth)


@step('I wait to see the value "{value}" in the "{nth:nth}" input "{name}"')
def wait_to_see_the_nth_input_with_value(ctx, value, nth, name):
    retry(assert_input_value)(ctx, name, value, index=nth)


@step('I should see no value in the "{nth:nth}" input "{name}"')
def should_see_the_nth_input_with_no_value(ctx, nth, name):
    assert_input_value(ctx, name, None, index=nth)
